#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on May 9, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy import uic
from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QListWidgetItem
from traits.api import Undefined

from karabo.common.api import Capabilities
from karabogui.singletons.api import get_topology

from .utils import get_dialog_ui


class DeviceCapabilityDialog(QDialog):
    def __init__(self, device_id="", capability=Capabilities.PROVIDES_SCENES,
                 parent=None):
        """A dialog to load capability items provided by devices

        :param device_id: If provided, the single device to request items from
        :param parent: The parent of the dialog
        """
        super().__init__(parent)
        filepath = get_dialog_ui("device_capability.ui")
        uic.loadUi(filepath, self)

        self.capability = capability
        # The currently selected device configuration
        self._selected_device = None
        self._requested_capa = None

        if device_id:
            # Add a single device and select it
            self.deviceNames.addItem(device_id)
            self.deviceNames.setCurrentItem(self.deviceNames.item(0))
            self.device_filter.setVisible(False)
            self.clear_button.setVisible(False)
        else:
            # Fill in the available devices
            self.topology = self._find_devices_with_capability()
            self.deviceNames.addItems(self.topology)

        # Initialize the Ok button
        self._enable_ok_button()

    @property
    def device_id(self):
        item = self.deviceNames.currentItem()
        if item:
            return item.text()
        return ""

    @property
    def capa_name(self):
        item = self.capaNames.currentItem()
        if item:
            return item.text()
        return ""

    def done(self, result):
        """Reimplement ``QDialog`` virtual slot"""
        self._clean_up()
        super().done(result)

    @Slot(str)
    def on_device_filter_textChanged(self, text):
        """Perform a case insensitive filtering on the topology"""
        topo = [device for device in self.topology if text.lower()
                in device.lower()]
        self.deviceNames.clearSelection()
        self.deviceNames.clear()
        self.deviceNames.addItems(topo)
        self.capaNames.clear()

    @Slot()
    def on_deviceNames_itemSelectionChanged(self):
        """A device was selected. Update the available capabilities.
        """
        self.capaNames.clear()

        item = self.deviceNames.currentItem()
        if item:
            provided_item = self._get_device_items(item.text())
            self.capaNames.addItems(provided_item)

    @Slot()
    def on_capaNames_itemSelectionChanged(self):
        """Synchronize the Ok button with the capability selection
        """
        self._enable_ok_button()

    @Slot(QListWidgetItem)
    def on_capaNames_itemDoubleClicked(self, item):
        """Just accept immediately when a capability is double-clicked"""
        self.accept()

    # -----------------------------------------------------------------------
    # Public interface

    def find_and_select(self, name):
        """Public interface to find and selected the capa name `name`"""
        self._requested_capa = name
        self._set_requested_capa()

    # -----------------------------------------------------------------------

    def _set_requested_capa(self):
        name = self._requested_capa
        if name is None:
            return

        items = self.capaNames.findItems(name, Qt.MatchExactly)
        if items:
            self.capaNames.setCurrentItem(items[0])
            self._requested_capa = None

    def _clean_up(self):
        """Clean up:
            - Trait notifications
            - Device monitor subscriptions
        """
        if self._selected_device is None:
            return

        device_proxy = self._selected_device
        device_proxy.on_trait_change(self._device_update, "config_update",
                                     remove=True)
        device_proxy.remove_monitor()
        self._selected_device = None

    def _enable_ok_button(self):
        """Only enable Ok button when a capability is selected
        """
        enabled = len(self.capaNames.selectedItems()) == 1
        self.buttonBox.button(QDialogButtonBox.Ok).setEnabled(enabled)

    def _find_devices_with_capability(self):
        """Walk the system topology and collect IDs of devices with capability
        on offer.
        """
        device_ids = []

        def _test_mask(mask, bit):
            return (mask & bit) == bit

        def visitor(node):
            nonlocal device_ids
            if _test_mask(node.capabilities, self.capability):
                device_ids.append(node.node_id)

        get_topology().visit_system_tree(visitor)
        device_ids.sort()

        return device_ids

    def _get_device_items(self, device_id):
        """Walk the system topology and collect IDs of devices with capability
        on offer.
        """
        device = get_topology().get_device(device_id)
        try:
            if self.capability is Capabilities.PROVIDES_SCENES:
                items = device.binding.value.availableScenes.value
            elif self.capability is Capabilities.PROVIDES_MACROS:
                items = device.binding.value.availableMacros.value
        except AttributeError:
            # This device lied about its capabilities or the information was
            # not yet received!
            binding = device.binding.value
            if not len(binding):
                self._request_device_capabilities(device)
            return []
        if items == [] or items is Undefined:
            # Not loaded yet. Request it.
            self._request_device_capabilities(device)
            # and return for now an empty list
            return []

        return items

    def _request_device_capabilities(self, device_proxy):
        """Request the configuration from a device so that we can get the
        available capabilities
        """
        self._clean_up()

        device_proxy.on_trait_change(self._device_update, "config_update")
        device_proxy.add_monitor()
        self._selected_device = device_proxy

    def _device_update(self, obj, name, new):
        if obj is not self._selected_device:
            return

        self._clean_up()
        self.on_deviceNames_itemSelectionChanged()
        self._set_requested_capa()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
import pytest

from karabogui.graph.image.api import bytescale, rescale

SIZE = 10
LOW, HIGH = (-10.0, 10.0)
PRECISION_TOLERANCE = 4


def test_rescale():
    """Test the rescale function of the image utils"""

    def _assert_dtype(dtype):
        """Check according to different `dtype`"""
        min_level, max_level = (0, 127)
        image = np.linspace(min_level, max_level, endpoint=True,
                            num=SIZE, dtype=dtype)

        # Check rescale with the low and high as the data extrema.
        # Note: It should not rescale!
        rescaled = rescale(image, min_level, max_level,
                           low=min_level, high=max_level)
        # Use `array_almost_equal` here to avoid precision inequality
        np.testing.assert_almost_equal(image, rescaled,
                                       decimal=PRECISION_TOLERANCE)

        # Check different low and high
        rescaled = rescale(image, min_level, max_level, low=LOW, high=HIGH)
        assert rescaled[0] == LOW
        assert rescaled[-1] == HIGH

    _assert_dtype(np.uint8)
    _assert_dtype(np.int8)
    _assert_dtype(np.uint32)
    _assert_dtype(np.int32)
    _assert_dtype(np.float32)


def test_bytescale():
    """Test the `bytescale` function of the image utils"""

    min_level, max_level = (0, 127)
    image = np.linspace(min_level, max_level, endpoint=True,
                        num=SIZE, dtype=np.uint8)

    bytescaled = bytescale(image, cmin=min_level, cmax=max_level,
                           low=min_level, high=max_level)
    np.testing.assert_almost_equal(image, bytescaled,
                                   decimal=PRECISION_TOLERANCE)

    bytescaled = bytescale(image, cmin=min_level, cmax=max_level,
                           low=5, high=100)
    assert bytescaled[0] == 5
    assert bytescaled[-1] == 100

    bytescaled = bytescale(image, cmin=None, cmax=None,
                           low=5, high=100)
    assert bytescaled[0] == 5
    assert bytescaled[-1] == 100

    image = np.linspace(min_level, max_level, endpoint=True,
                        num=SIZE, dtype=np.uint32)
    bytescaled = bytescale(image, cmin=min_level, cmax=max_level,
                           low=5, high=100)
    assert bytescaled[0] == 5
    assert bytescaled[-1] == 100

    image = np.linspace(min_level, max_level, endpoint=True,
                        num=SIZE, dtype=np.float32)
    bytescaled = bytescale(image, cmin=min_level, cmax=max_level,
                           low=5, high=100)
    assert bytescaled[0] == 5
    assert bytescaled[-1] == 100

    with pytest.raises(ValueError):
        # No negative low allowed
        bytescale(image, cmin=min_level, cmax=max_level,
                  low=-5, high=100)

    with pytest.raises(ValueError):
        # High must be higher than low
        bytescale(image, cmin=min_level, cmax=max_level,
                  low=127, high=100)



# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.

from karabogui.testing import singletons, system_hash
from karabogui.topology.api import SystemTopology

from ..server_handle import ServerHandleDialog


def test_server_handle_dialog(gui_app):
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(topology=topology):
        dialog = ServerHandleDialog()
        assert dialog.ui_server_id.count() == 1
        assert dialog.server_id == "swerver"
        dialog.done(0)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QModelIndex, QPoint, QSize, Qt, Slot
from qtpy.QtWidgets import QDialog

from karabo.native import AccessLevel, Hash
from karabogui.events import KaraboEvent
from karabogui.navigation.system_model import SystemTreeModel
from karabogui.navigation.system_view import SystemTreeView
from karabogui.singletons.mediator import Mediator
from karabogui.testing import GuiTestCase, singletons, system_hash


class TestCase(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.view = SystemTreeView()
        self.model = self.view.model()
        self.system_model = self.model.sourceModel()
        tree = self.system_model.tree
        tree.initialize(system_hash())

    def tearDown(self):
        super().tearDown()
        self.view.destroy()
        self.model = None
        self.system_model = None
        self.view = None

    def test_view_click(self):
        self.assertTrue(self.view.expanded)
        self.view.onDoubleClickHeader()
        self.assertFalse(self.view.expanded)
        self.view.onDoubleClickHeader()
        self.assertTrue(self.view.expanded)

        # System model view is expanded on reset
        self.view.resetExpand()
        self.assertTrue(self.view.expanded)

        self.view.expandAll()
        self.assertTrue(self.view.expanded)
        self.view.collapseAll()
        self.assertFalse(self.view.expanded)

    def test_index_actions(self):
        selection_tracker = mock.Mock()
        model = self.model
        host_index = model.index(0, 0)
        server_index = model.index(0, 0, host_index)
        class_index = model.index(0, 0, server_index)
        device_index = model.index(0, 0, class_index)
        self.assertEqual(device_index.data(), "divvy")

        deviceId = ""
        navigation_type = ""
        device_proxy = None

        assert host_index.data() == "BIG_IRON"
        with singletons(mediator=Mediator()):
            call_slot = "karabogui.navigation.system_view.call_device_slot"
            is_online = "karabogui.navigation.system_view.is_device_online"
            with mock.patch(is_online) as ol:
                ol.return_value = True
                with mock.patch(call_slot) as call:
                    self.view.setCurrentIndex(host_index)
                    self.view.onNetworkInfo()
                    args = call.call_args.args
                    kwargs = call.call_args.kwargs
                    assert "BIG_IRON" in kwargs.values()
                    assert "requestNetwork" in args

        @Slot()
        def receiveSignal(nav_type, proxy):
            nonlocal navigation_type, deviceId, device_proxy
            navigation_type = nav_type
            deviceId = proxy.device_id if proxy is not None else None
            device_proxy = proxy

        model.signalItemChanged.connect(receiveSignal)

        index = device_index
        broadcast_path = "karabogui.navigation.system_view.broadcast_event"
        # Protect mediator in tests
        with singletons(selection_tracker=selection_tracker):
            with mock.patch(broadcast_path) as broadcast:
                model.selectIndex(index)
                self.process_qt_events()
                self.assertEqual(navigation_type, "device")
                self.assertEqual(deviceId, "divvy")
                broadcast.assert_called_with(KaraboEvent.ShowConfiguration,
                                             {"proxy": device_proxy})

        self.assertEqual(self.view.currentIndex(), index)
        self.assertTrue(self.view.currentIndex().isValid())
        # Check indexInfo
        info = self.view.indexInfo()
        self.assertEqual(info["deviceId"], "divvy")

        popup_widget = "karabogui.navigation.system_view.PopupWidget"
        with mock.patch(popup_widget) as widget:
            self.view.onAbout()
            widget().setInfo.assert_called_with(
                {"host": "BIG_IRON",
                 "archive": True,
                 "type": "device",
                 "visibility": AccessLevel.OBSERVER,
                 "capabilities": 5,
                 "interfaces": 0,
                 "serverId": "swerver",
                 "classId": "FooClass",
                 "status": "ok"})

        conf_widget = "karabogui.navigation.system_view." \
                      "ConfigurationFromPastDialog"
        with mock.patch(conf_widget) as widget:
            self.view.onGetConfigurationFromPast()
            widget.assert_called_once()

        conf_widget = "karabogui.navigation.system_view." \
                      "InitConfigurationDialog"
        with mock.patch(conf_widget) as widget:
            self.view.onGetInitConfiguration()
            widget.assert_called_once()

        manager = mock.Mock()
        with singletons(manager=manager):
            with mock.patch("karabogui.singletons.manager.QMessageBox") as mb:
                mb().size.return_value = QSize(10, 10)
                mb().exec.return_value = mb.Yes
                self.view.onKillInstance()
                manager.shutdownDevice.assert_called_once()

        manager.reset_mock()
        with singletons(manager=manager):
            scene_capa = "karabogui.navigation.system_view." \
                         "DeviceCapabilityDialog"
            with mock.patch(scene_capa) as scene:
                scene().device_id = "divvy"
                scene().capa_name = "scene"
                scene().exec.return_value = QDialog.Accepted
                self.view.onOpenDeviceScene()
                args = manager.callDeviceSlot.call_args[0]
                self.assertIn("requestScene", args)
                self.assertIn("divvy", args)
                manager.callDeviceSlot.assert_called_once()

        manager.reset_mock()
        with singletons(manager=manager):
            self.view.onTimeInformation()
            args = manager.callDeviceSlot.call_args[0]
            self.assertIn("divvy", args)
            self.assertIn("slotGetTime", args)
            manager.callDeviceSlot.assert_called_once()

        with singletons(selection_tracker=selection_tracker):
            with mock.patch(broadcast_path):
                model.selectIndex(server_index)

        manager.reset_mock()
        with singletons(manager=manager):
            with mock.patch("karabogui.singletons.manager.QMessageBox") as mb:
                mb().size.return_value = QSize(10, 10)
                mb().exec.return_value = mb.Yes
                self.view.onKillInstance()
                manager.shutdownServer.assert_called_once()
                # Servers don't have proxies
                self.assertEqual(device_proxy, None)

        network = mock.Mock()
        with singletons(network=network):
            path = "karabogui.navigation.system_view.QInputDialog"
            with mock.patch(path) as dia:
                dia.getItem.return_value = "DEBUG", True
                self.view.onLoggerLevel()
                network.onSetLogLevel.assert_called_with("swerver", "DEBUG")

    def test_context_menu_device_view(self):
        selection_tracker = mock.Mock()
        pos = QPoint(0, 0)
        menu_path = "karabogui.navigation.system_view.QMenu"
        with mock.patch(menu_path) as menu:
            view = SystemTreeView()
            model = view.model()
            host_index = model.index(0, 0)
            server_index = model.index(0, 0, host_index)
            class_index = model.index(0, 0, server_index)
            index = model.index(0, 0, class_index)
            broadcast_path = "karabogui.navigation.system_view.broadcast_event"
            # Protect mediator in tests
            with singletons(selection_tracker=selection_tracker):
                with mock.patch(broadcast_path):
                    model.selectIndex(index)
            view.onCustomContextMenuRequested(pos)
            menu().exec.assert_called_once()
            view.destroy()

    def test_drag_actions(self):
        self.assertEqual(self.system_model.supportedDragActions(),
                         Qt.CopyAction)

    def test_header(self):
        header = self.system_model.headerData(0, Qt.Horizontal, Qt.DisplayRole)
        assert header == "Host - Server - Class - Device"

    def test_columnCount_root(self):
        self.assertEqual(self.system_model.columnCount(), 2)

    def test_columnCount_filter_model(self):
        self.assertEqual(self.model.columnCount(), 2)

    def test_rowCount_filter_model(self):
        self.assertEqual(self.model.rowCount(), 1)

    def test_rowCount_root(self):
        self.assertEqual(self.system_model.rowCount(), 1)

    def test_dynamics_sorting(self):
        assert not self.model.dynamicSortFilter()

    def test_update_topo(self):
        """Add a device to the system model"""

        host_index = self.system_model.index(0, 0)
        assert host_index.isValid()
        assert host_index.data() == "BIG_IRON"

        host = host_index.internalPointer()
        # We have 1 server on host
        assert len(host.children) == 1

        server_index = self.system_model.index(0, 0, host_index)
        server = server_index.internalPointer()
        # We have 2 classes
        assert len(server.children) == 2

        class_index = self.system_model.index(0, 0, server_index)
        class_node = class_index.internalPointer()
        assert len(class_node.children) == 1

        # We only have 1 device
        device_index = self.system_model.index(0, 0, class_index)
        assert device_index.data() == "divvy"

        # We add a single device with same class
        h = Hash()
        h["device.newdevice"] = None
        h["device.newdevice", ...] = {
            "host": "BIG_IRON",
            "visibility": AccessLevel.OBSERVER,
            "serverId": "swerver",
            "classId": "FooClass",
            "status": "ok",
            "capabilities": 0,
        }
        self.system_model.tree.update(h)
        # __none__ server is not accounted anymore ...
        assert self.system_model.rowCount() == 1

        host_index = self.system_model.index(0, 0)
        assert host_index.isValid()
        assert host_index.data() == "BIG_IRON"

        host = host_index.internalPointer()
        assert len(host.children) == 1

        # servers
        server_index = self.system_model.index(0, 0, host_index)
        server = server_index.internalPointer()
        assert len(server.children) == 2

        # We now have two devices
        class_index = self.system_model.index(0, 0, server_index)
        class_node = class_index.internalPointer()
        assert len(class_node.children) == 2

        device_index = self.system_model.index(0, 0, class_index)
        assert device_index.data() == "divvy"

        device_index = self.system_model.index(1, 0, class_index)
        assert device_index.data() == "newdevice"

    def test_update_topo_remove_device(self):
        """Remove a device from the system model"""
        self.system_model.tree.remove_device("newdevice")

        host_index = self.system_model.index(0, 0)
        assert host_index.data() == "BIG_IRON"

        host = host_index.internalPointer()
        assert len(host.children) == 1

        # servers
        server_index = self.system_model.index(0, 0, host_index)
        server = server_index.internalPointer()
        assert len(server.children) == 2

        # We removed the added device
        class_index = self.system_model.index(0, 0, server_index)
        assert class_index.data() == "FooClass"
        class_node = class_index.internalPointer()
        assert len(class_node.children) == 1

        device_index = self.system_model.index(0, 0, class_index)
        assert device_index.data() == "divvy"

    def test_find_nodes(self):
        """Test that we can find a node"""
        node = self.model.findNodes("divvy")
        assert len(node)

    # ------------------------------------------------------------
    # Filter test

    def test_filter_model_device(self):
        """Test that we can filter in the system model"""
        self.model.setFilterFixedString("divvy")
        # Check that after correct filtering the device is there
        host_index = self.model.index(0, 0)
        assert host_index.isValid()
        assert host_index.data() == "BIG_IRON"
        server_index = self.model.index(0, 0, host_index)
        class_index = self.model.index(0, 0, server_index)
        device_index = self.model.index(0, 0, class_index)
        assert device_index.data() == "divvy"

        self.model.setFilterFixedString("xfel")
        host_index = self.model.index(0, 0)
        # Host is not valid, filtered out
        assert not host_index.isValid()

    # ------------------------------------------------------------
    # Basic model stress test

    def test_data_bad_index(self):
        self.assertIsNone(self.system_model.data(QModelIndex(),
                                                 Qt.DisplayRole))

    def test_modeltester_qt(self):
        from pytestqt.modeltest import ModelTester
        system_model = SystemTreeModel()
        tree = system_model.tree
        tree.initialize(system_hash())
        tester = ModelTester(None)
        tester.check(system_model)


if __name__ == "__main__":
    main()



import pytest
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialogButtonBox, QMessageBox

from karabogui.graph.plots.api import CurveOptionsDialog
from karabogui.testing import click_button

PINK = "#fb9a99"
ORANGE = "#ff7f00"

curve1 = object()
curve2 = object()

OPTIONS = {curve1: {"key": "first_plot", "pen_color": ORANGE,
                    "name": "Curve 1", "curve_type": 1},
           curve2: {"key": "second_plot", "pen_color": PINK,
                    "name": "Curve 2", "curve_type": 1}}


@pytest.fixture
def dialog(gui_app):
    dialog = CurveOptionsDialog(curve_options=OPTIONS)
    yield dialog


def test_load_options(dialog):
    proxy_list = dialog.proxy_list
    assert proxy_list.count() == len(OPTIONS)
    assert proxy_list.item(0).text() == "first_plot"
    assert proxy_list.item(0).data(Qt.UserRole) == OPTIONS[curve1]
    assert proxy_list.item(0).data(Qt.UserRole + 1) == curve1

    assert proxy_list.item(1).text() == "second_plot"
    assert proxy_list.item(1).data(Qt.UserRole) == OPTIONS[curve2]
    assert proxy_list.item(1).data(Qt.UserRole + 1) == curve2


def test_color_combobox(dialog):
    assert dialog.color_combo_box.count() == 13
    assert dialog.color_combo_box.itemText(0) == ""
    assert dialog.color_combo_box.itemText(11) == ""


def test_displayed_options(dialog):
    dialog.proxy_list.setCurrentRow(0)
    dialog.color_combo_box.currentText() == ""
    dialog.legend_name.text() == "Curve 1"
    dialog.property_key.text() == "first_plot"

    dialog.proxy_list.setCurrentRow(1)
    dialog.color_combo_box.currentText() == ""
    dialog.legend_name.text() == "Curve 2"
    dialog.property_key.text() == "second_plot"


def test_get_curve_options(dialog):
    # No changes yet
    assert dialog.get_curve_options() == {}
    assert dialog._original == OPTIONS

    # Change option values.
    dialog.proxy_list.setCurrentRow(1)
    dialog.legend_name.setText("New Legend")
    dialog.legend_name.editingFinished.emit()
    dialog.color_combo_box.setCurrentText("Orange")
    curve_options = dialog.get_curve_options()
    assert curve_options[curve1] == OPTIONS[curve1]

    assert curve_options[curve2]["pen_color"] == PINK
    assert curve_options[curve2]["name"] == "New Legend"


def test_get_curve_options_reset(dialog, mocker):
    path = "karabogui.graph.plots.dialogs.curve_options.QMessageBox"
    message_box = mocker.patch(path)
    message_box.return_value = QMessageBox.Yes
    click_button(dialog.buttonBox.button(QDialogButtonBox.Reset))
    assert dialog.has_reset
    assert dialog.get_curve_options() is None



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 23, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import QByteArray, QPoint, QRect, Qt
from qtpy.QtGui import QBrush, QPainter, QPen
from qtpy.QtSvg import QSvgRenderer
from qtpy.QtWidgets import QWidget

from karabo.common.scenemodel.api import write_single_model
from karabogui.widgets.hints import KaraboSceneWidget


class UnknownSvgWidget(QWidget):
    """ A widget which can display data from an UnknownXMLDataModel.
    """

    def __init__(self, renderer, parent=None):
        super().__init__(parent)
        self.renderer = renderer
        self.setGeometry(renderer.viewBox())
        self.setToolTip("External SVG Element")

    def minimumSizeHint(self):
        return self.renderer.defaultSize()

    def paintEvent(self, event):
        with QPainter(self) as painter:
            self.renderer.render(painter)

    def sizeHint(self):
        return self.minimumSizeHint()

    @classmethod
    def create(cls, model, parent=None):
        """Create an instance of this widget from a model object.

        Returns None if there would be nothing to display.
        """
        xml = write_single_model(model).encode("utf-8")
        ar = QByteArray.fromRawData(xml)
        renderer = QSvgRenderer(ar)

        if renderer.isValid() and not renderer.defaultSize().isNull():
            return cls(renderer, parent=parent)
        return None

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def set_geometry(self, rect):
        """Satisfy the informal widget interface."""

    def translate(self, offset):
        """Satisfy the informal widget interface."""


class UnknownWidget(KaraboSceneWidget, QWidget):
    """A widget which can display data from an UnknownWidgetDataModel or from a
    BaseWidgetObjectData subclass which is unsupported for some reason.
    """

    def __init__(self, model, parent=None):
        super().__init__(model=model, parent=parent)
        klassname = type(self.model).__name__
        if hasattr(self.model, "klass"):
            klassname = self.model.klass
        keys = ", ".join(self.model.keys)
        self.setToolTip(f"{keys}: --- Unsupported widget type: {klassname}")
        self.setGeometry(QRect(model.x, model.y, model.width, model.height))

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_alarm(self):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""

    def set_geometry(self, rect):
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    def paintEvent(self, event):
        with QPainter(self) as painter:
            unsupported_text = "Unsupported"

            rect = self.rect()
            boundary = rect.adjusted(2, 2, -2, -2)
            painter.fillRect(rect, QBrush(Qt.lightGray, Qt.FDiagPattern))

            pen = QPen(Qt.lightGray)
            pen.setJoinStyle(Qt.MiterJoin)
            pen.setWidth(4)
            painter.setPen(pen)
            painter.drawRect(boundary)

            metrics = painter.fontMetrics()
            text_rect = metrics.boundingRect(unsupported_text)
            pos = rect.center() - QPoint(text_rect.width() // 2,
                                         -text_rect.height() // 2)
            painter.setPen(QPen())
            painter.drawText(pos, unsupported_text)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import QPoint

from karabo.common.scenemodel.api import TableElementModel
from karabo.native import Bool, Configurable, Hash, String, VectorHash
from karabogui.binding.config import apply_configuration
from karabogui.controllers.table.api import BaseTableController, TableModel
from karabogui.testing import get_property_proxy


class TableSchema(Configurable):
    arch = String(
        defaultValue="NoString")
    foo = Bool(
        defaultValue=False,
        displayedName="Foo")


class Object(Configurable):
    prop = VectorHash(rows=TableSchema)


class MyTableModel(TableModel):
    state_less = True


count = 0
filter_model = False
legacy_model = False


class MyTableController(BaseTableController):
    # A table model without resizeToContents
    hasCustomMenu = True
    tableModelClass = MyTableModel

    def custom_menu(self, pos):
        global count
        count += 1

    def createFilterModel(self, item_model):
        global filter_model
        filter_model = True
        return item_model


class LegacyTableController(BaseTableController):
    tableModelClass = MyTableModel

    def createModel(self, item_model):
        global legacy_model
        legacy_model = True
        return item_model


@pytest.fixture
def custom_table_controller_setup(gui_app):
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = TableElementModel()
    controller = MyTableController(proxy=proxy, model=model)
    controller.create(None)
    assert controller.isReadOnly()
    controller.set_read_only(False)
    table_hash = Hash(
        "prop",
        [Hash("arch", "a", "foo", True),
         Hash("arch", "b", "foo", False)])
    apply_configuration(table_hash, proxy.root_proxy.binding)

    yield controller

    controller.destroy()
    assert controller.widget is None


def test_subclassing_model(custom_table_controller_setup):
    """Test that one can specify the table model"""
    controller = custom_table_controller_setup
    model = controller.sourceModel()
    assert model.state_less
    assert isinstance(model, MyTableModel)
    assert filter_model is True


def test_custom_context(custom_table_controller_setup):
    """Test that one can use a custom context menu"""
    controller = custom_table_controller_setup
    # Note: Try QTest on later versions with rightclick, not working in
    # Qt 5.9
    assert count == 0
    controller._custom_menu(QPoint(0, 0))
    # Calls custom_menu
    assert count == 1


def test_column_index_key(custom_table_controller_setup):
    controller = custom_table_controller_setup
    index = controller.columnIndex("foo")
    assert index == 1
    index = controller.columnIndex("arch")
    assert index == 0
    # Schema evolution! Return `None` index
    index = controller.columnIndex("nothere")
    assert index is None

    key = controller.columnKey(0)
    assert key == "arch"
    key = controller.columnKey(1)
    assert key == "foo"
    key = controller.columnKey(2)
    assert key is None


def test_stretch_last_section(custom_table_controller_setup):
    controller = custom_table_controller_setup
    table_widget = controller.tableWidget()
    assert table_widget.horizontalHeader().stretchLastSection()


def test_legacy_controller(gui_app):
    proxy = get_property_proxy(Object.getClassSchema(), "prop")
    model = TableElementModel()
    controller = LegacyTableController(proxy=proxy, model=model)
    controller.create(None)
    assert controller.isReadOnly()
    assert legacy_model

    # Check the menu
    menu = controller.get_basic_menu()
    assert menu is not None
    # Readonly does not have an action
    assert len(menu.actions()) == 0
    controller.set_read_only(False)
    menu = controller.get_basic_menu()
    assert menu is not None
    assert len(menu.actions()) == 1



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import mock

from qtpy.QtWidgets import QLabel
from traits.api import Dict, Instance, Int, Str, Undefined

from karabo.common.api import KARABO_SCHEMA_DISPLAYED_NAME, State
from karabo.common.scenemodel.api import BaseWidgetObjectData
from karabo.native import AccessMode, Bool, Configurable, Hash, Node, String
from karabogui.binding.api import (
    DeviceClassProxy, NodeBinding, PropertyProxy, ProxyStatus, StringBinding,
    apply_configuration, build_binding, get_binding_value)
from karabogui.testing import (
    GuiTestCase, flushed_registry, get_property_proxy, set_proxy_value)

from ..base import BaseBindingController
from ..registry import register_binding_controller
from ..util import with_display_type


class SampleObject(Configurable):
    first = String(accessMode=AccessMode.READONLY, displayedName='First:')
    second = String(accessMode=AccessMode.READONLY)
    unsupported = Bool(accessMode=AccessMode.READONLY)


class ChangeObject(Configurable):
    first = String(accessMode=AccessMode.READONLY, displayedName='Change:')


class NodedObject(Configurable):
    node = Node(SampleObject)


class StateObject(Configurable):
    state = String(enum=State, displayType='State')
    other = String()


# Classes for testing late binding initialization
class Unconnected(Configurable):
    pass


class Connected(Configurable):
    one = String(displayedName='One:')
    two = String(displayedName='Two:')


class UniqueWidgetModel(BaseWidgetObjectData):
    pass  # Satisfy the uniqueness check in register_binding_controller


def _define_binding_classes():
    """Do this in a function so that we can avoid polluting the registry
    """

    @register_binding_controller(klassname='Norm', binding_type=StringBinding)
    class DeviceController(BaseBindingController):
        model = Instance(UniqueWidgetModel, args=())
        display_names = Dict

        def add_proxy(self, proxy):
            self._set_display_name(proxy)
            return True

        def binding_update(self, proxy):
            self._set_display_name(proxy)

        def create_widget(self, parent):
            return QLabel(parent)

        def _set_display_name(self, proxy):
            binding = proxy.binding
            if binding is not None:
                name = binding.attributes.get(KARABO_SCHEMA_DISPLAYED_NAME, '')
            else:
                name = ''
            self.display_names[proxy] = name

    @register_binding_controller(klassname='Multi', binding_type=StringBinding,
                                 can_show_nothing=True)
    class MultiBindingController(BaseBindingController):
        model = Instance(UniqueWidgetModel, args=())

        def add_proxy(self, proxy):
            return True

        def remove_proxy(self, proxy):
            return True

        def binding_update(self, proxy):
            return True

        def create_widget(self, parent):
            return QLabel(parent)

        def value_update(self, proxy):
            value = get_binding_value(proxy)
            if value is not None:
                self.widget.setText(value)

    @register_binding_controller(klassname="Don't show nothing",
                                 binding_type=StringBinding,
                                 can_show_nothing=False)
    class DontShowNothingBindingController(BaseBindingController):
        model = Instance(UniqueWidgetModel, args=())

        def add_proxy(self, proxy):
            return True

        def binding_update(self, proxy):
            return True

        def create_widget(self, parent):
            return QLabel(parent)

        def value_update(self, proxy):
            self.widget.setText(proxy.value)

    @register_binding_controller(klassname='Node', binding_type=NodeBinding)
    class NodeBindingController(BaseBindingController):
        model = Instance(UniqueWidgetModel, args=())

        def create_widget(self, parent):
            return QLabel(parent)

        def binding_update(self, proxy):
            return True

        def value_update(self, proxy):
            self.widget.setText(proxy.binding.value.second.value)

    @register_binding_controller(klassname='Mono', binding_type=StringBinding)
    class SingleBindingController(BaseBindingController):
        model = Instance(UniqueWidgetModel, args=())
        disp_name = Str
        deferred = Int(0)

        def binding_update(self, proxy):
            attrs = proxy.binding.attributes
            self.disp_name = attrs.get(KARABO_SCHEMA_DISPLAYED_NAME)

        def create_widget(self, parent):
            return QLabel(parent)

        def deferred_update(self):
            self.deferred += 1

        def value_update(self, proxy):
            value = get_binding_value(proxy)
            if value is not None:
                self.widget.setText(self.disp_name + proxy.value)

    @register_binding_controller(klassname='State', binding_type=StringBinding,
                                 is_compatible=with_display_type('State'))
    class StateTrackingController(BaseBindingController):
        model = Instance(UniqueWidgetModel, args=())

        def create_widget(self, parent):
            return QLabel(parent)

        def state_update(self, proxy):
            self.widget.setText(proxy.root_proxy.state_binding.value)

    return {
        'DeviceController': DeviceController,
        'MultiBindingController': MultiBindingController,
        'DontShowNothingBindingController': DontShowNothingBindingController,
        'NodeBindingController': NodeBindingController,
        'SingleBindingController': SingleBindingController,
        'StateTrackingController': StateTrackingController,
    }


class TestBaseBindingController(GuiTestCase):
    def setUp(self):
        super().setUp()

        # Avoid polluting the controller registry!
        with flushed_registry():
            klasses = _define_binding_classes()

        # Save the classes so they only need to be declared once
        self.DeviceController = klasses['DeviceController']
        self.MultiBindingController = klasses['MultiBindingController']
        self.NodeBindingController = klasses['NodeBindingController']
        self.SingleBindingController = klasses['SingleBindingController']
        self.StateTrackingController = klasses['StateTrackingController']
        self.DontShowNothingBindingController = \
            klasses['DontShowNothingBindingController']

        schema = SampleObject.getClassSchema()
        binding = build_binding(schema)
        device = DeviceClassProxy(binding=binding, server_id='Fake',
                                  status=ProxyStatus.OFFLINE)
        self.first = PropertyProxy(root_proxy=device, path='first')
        self.second = PropertyProxy(root_proxy=device, path='second')
        self.unsupported = PropertyProxy(root_proxy=device, path='unsupported')

        # Create the controllers and initialize their widgets
        self.single = self.SingleBindingController(proxy=self.first)
        self.single.create(None)
        self.single.finish_initialization()

        # The order of initializing a controller with multiple properties are
        # as follows:
        # 1. instantiate controller class
        # 2. controller.create(parent) with parent of QWidget type or None
        # 3. controller.visualize_additional_property(proxy) for addtl proxies
        # 4. controller.finish_initialization()
        self.multi = self.MultiBindingController(proxy=self.first)
        self.multi.create(None)
        assert self.multi.visualize_additional_property(self.second)
        self.multi.finish_initialization()

        # Create a mock  slot
        self.received_values = []

    def tearDown(self):
        self.single.destroy()
        self.multi.destroy()
        assert self.single.widget is None
        assert self.multi.widget is None

    def test_extra_properties(self):
        # SingleBindingController doesn't support multiple properties
        assert not self.single.visualize_additional_property(self.second)

        # Use a temporary controller which won't affect other tests
        model = UniqueWidgetModel(keys=['first'])
        with mock.patch.object(self.MultiBindingController, "binding_update") \
                as binding_update:
            controller = self.MultiBindingController(proxy=self.first,
                                                     model=model)
            controller.create(None)
            # Adding an already watched property doesn't work
            assert not controller.visualize_additional_property(self.first)
            assert controller.model.keys == ['first']

            binding_update.assert_called_once()
            binding_update.reset_mock()

            assert controller.visualize_additional_property(self.second)
            assert controller.model.keys == ['first', 'second']

            binding_update.assert_not_called()

            assert not controller.visualize_additional_property(
                self.second)
            assert not controller.visualize_additional_property(
                self.unsupported)
            assert controller.model.keys == ['first', 'second']
            binding_update.assert_not_called()

            # Remove an additional proxy
            assert self.second in controller.proxies
            controller.remove_additional_property(self.second)
            binding_update.assert_not_called()

            assert controller.model.keys == ['first']
            assert self.second not in controller.proxies

            # Already removed!
            assert not controller.remove_additional_property(self.second)

            # Value update is not effective anymore
            set_proxy_value(self.second, self.second.path, "NoUpdateShould")
            assert controller.widget.text() != "NoUpdateShould"

    def test_single_value_update(self):
        set_proxy_value(self.first, 'first', 'Foo')
        assert self.single.widget.text() == 'First:Foo'

    def test_multi_value_update(self):
        set_proxy_value(self.first, 'first', 'Bar')
        assert self.multi.widget.text() == 'Bar'
        set_proxy_value(self.second, 'second', 'Qux')
        assert self.multi.widget.text() == 'Qux'

    def test_node_value_update(self):
        binding = build_binding(NodedObject.getClassSchema())
        device = DeviceClassProxy(binding=binding, server_id='Test',
                                  status=ProxyStatus.OFFLINE)
        proxy = PropertyProxy(root_proxy=device, path='node')
        controller = self.NodeBindingController(proxy=proxy)
        controller.create(None)

        config = Hash('node', Hash('first', 'foo', 'second', 'bar'))
        apply_configuration(config, binding)
        assert controller.widget.text() == 'bar'

    def test_mocked_single_value_update(self):
        with mock.patch.object(self.SingleBindingController, "value_update",
                               side_effect=self._mocked_value_update) \
                as mocked_method:
            self._assert_value_update(proxy=self.first,
                                      value="Foo",
                                      value_update=mocked_method)

    def test_mocked_multi_value_update(self):
        with mock.patch.object(self.MultiBindingController, "value_update",
                               side_effect=self._mocked_value_update) \
                as mocked_method:
            self._assert_value_update(proxy=self.first,
                                      value="Bar",
                                      value_update=mocked_method)

            self._assert_value_update(proxy=self.second,
                                      value="Qux",
                                      value_update=mocked_method)

    def test_mocked_node_value_update(self):
        # Create a mocked value_update
        first_value, second_value = None, None

        def mocked_value_update(proxy):
            nonlocal first_value, second_value
            first_value = proxy.value.first.value
            second_value = proxy.value.second.value

        # Instantiate the node proxy and controller
        binding = build_binding(NodedObject.getClassSchema())
        device = DeviceClassProxy(binding=binding, server_id='Test',
                                  status=ProxyStatus.OFFLINE)
        proxy = PropertyProxy(root_proxy=device, path='node')
        controller = self.NodeBindingController(proxy=proxy)
        controller.create(None)

        # Trigger the value update
        with mock.patch.object(self.NodeBindingController, "value_update",
                               side_effect=mocked_value_update) \
                as value_update:
            config = Hash('node', Hash('first', 'foo', 'second', 'bar'))
            apply_configuration(config, binding)
            value_update.assert_called_once_with(proxy)
            assert first_value == 'foo'
            assert second_value == 'bar'

    def _mocked_value_update(self, proxy):
        self.received_values.append(proxy.value)

    def _assert_value_update(self, proxy, value, value_update):
        # Prepare the test properties
        value_update.reset_mock()
        self.received_values.clear()

        # Trigger value_update
        set_proxy_value(proxy, proxy.path, value)
        value_update.assert_called_once_with(proxy)
        assert self.received_values == [value]

    def test_init(self):
        # Check controller that can_show_nothing and without values
        self._assert_init_value_update(self.MultiBindingController,
                                       values=[Undefined, Undefined])

        # Check controller with can_show_nothing is False and without values
        self._assert_init_value_update(self.DontShowNothingBindingController,
                                       called=False)

        # Check controller with can_show_nothing is False and with proxies
        # having previous values
        first_value, second_value = 'Foo', 'Bar'
        set_proxy_value(self.first, self.first.path, first_value)
        set_proxy_value(self.second, self.second.path, second_value)
        self._assert_init_value_update(self.DontShowNothingBindingController,
                                       values=[second_value, first_value])

    def _assert_init_value_update(self, klass, called=True, values=None):
        """The initialization of the controller in the scene are done
        as follows:

        1. controller_klass(proxy, model) with the first proxy and model
        2. controller.create(parent) with parent as the container widget
        3. controller.visualize_additional_property(proxy) for every
           additional proxies
        4. controller.finish_initialization()
        """
        # Prepare the test properties
        self.received_values = []

        # Initialize the controller
        with mock.patch.object(klass, "value_update",
                               side_effect=self._mocked_value_update) \
                as value_update:
            controller = klass(proxy=self.first)
            controller.create(None)
            controller.visualize_additional_property(proxy=self.second)
            controller.finish_initialization()

        if called is False:
            value_update.assert_not_called()
        else:
            # We expect that the value_update is triggered by
            # additional proxies first, then the main proxy
            proxies = [self.second, self.first]
            calls = [mock.call(proxy) for proxy in proxies]
            value_update.assert_has_calls(calls)
            values = [proxy.value for proxy in proxies]
            assert self.received_values == values

    def test_deferred_update(self):
        deferred_before = self.single.deferred

        self.single.update_later()
        # Force the event loop to run so that the callback gets called
        self.process_qt_events()

        assert deferred_before + 1 == self.single.deferred

    def test_schema_injection(self):
        set_proxy_value(self.first, 'first', 'Now')
        assert self.single.widget.text() == 'First:Now'

        try:
            # New device schema should cause a call to
            # SingleBindingController._binding_update
            schema = ChangeObject.getClassSchema()
            build_binding(schema, existing=self.first.root_proxy.binding)
            # Schema injection resets the value!
            set_proxy_value(self.first, 'first', 'Now')
            assert self.single.widget.text() == 'Change:Now'

        finally:
            # Put things back as they were!
            schema = SampleObject.getClassSchema()
            build_binding(schema, existing=self.first.root_proxy.binding)

    def test_device_wakeup(self):
        # Build a widget with two proxies that have no `binding`
        binding = build_binding(Unconnected.getClassSchema())
        device = DeviceClassProxy(binding=binding, server_id='Test',
                                  status=ProxyStatus.OFFLINE)
        one = PropertyProxy(root_proxy=device, path='one')
        two = PropertyProxy(root_proxy=device, path='two')
        controller = self.DeviceController(proxy=one)
        controller.create(None)

        controller.visualize_additional_property(two)
        assert two in controller.display_names

        # Then swap the schema so that the bindings are now available
        build_binding(Connected.getClassSchema(), existing=device.binding)
        assert controller.display_names[one] == 'One:'
        assert controller.display_names[two] == 'Two:'

    def test_device_state_tracking(self):
        binding = build_binding(StateObject.getClassSchema())
        device = DeviceClassProxy(binding=binding, server_id='Test',
                                  status=ProxyStatus.OFFLINE)
        # Make a proxy for something other than the 'state' property
        proxy = PropertyProxy(root_proxy=device, path='other')
        controller = self.StateTrackingController(proxy=proxy)
        controller.create(None)

        set_proxy_value(proxy, 'state', 'ERROR')
        assert controller.widget.text() == 'ERROR'

    def test_device_state_bad_binding(self):
        binding = build_binding(StateObject.getClassSchema())
        device = DeviceClassProxy(binding=binding, server_id='Test',
                                  status=ProxyStatus.OFFLINE)
        # Make a proxy for something which doesn't exist
        proxy = PropertyProxy(root_proxy=device, path='oaoijasdoijasd')
        controller = self.StateTrackingController(proxy=proxy)
        controller.create(None)

        set_proxy_value(proxy, 'state', 'ERROR')
        assert controller.widget.text() == ''

    def test_get_instance_id(self):
        proxy = get_property_proxy(None, "prop")
        controller = self.StateTrackingController(proxy=proxy)
        assert controller.getInstanceId() == "TestDevice"



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import numpy as np
from qtpy.QtCore import QObject, QPointF, QRectF, Qt, Slot
from qtpy.QtGui import QColor, QPen
from qtpy.QtWidgets import QGraphicsObject, QGraphicsRectItem

from karabogui.graph.common.api import CoordsLegend, float_to_string
from karabogui.graph.common.const import TF_SCALING, TF_TRANSLATION
from karabogui.graph.image.legends.picker import PickerLegend


class PickerController(QObject):

    def __init__(self, plotItem, parent=None):
        super().__init__(parent=parent)

        self.plotItem = plotItem
        self._selected_pixel = None

        # Indicator rectangle for hovered pixel.
        # Only appears when image is hovered
        self._indicator_rect = IndicatorRectangle()
        self._indicator_rect.setZValue(10)
        self.plotItem.vb.addItem(self._indicator_rect,
                                 ignoreBounds=True)

        # Selection rectange for clicked pixel. Appears when image is clicked.
        # Persistent when re-selecting the Picker tool button.
        self._selection_rect = IndicatorRectangle()
        self._selection_rect.setZValue(10)
        self.plotItem.vb.addItem(self._selection_rect,
                                 ignoreBounds=True)

        # Coords legend for hovered pixel coords. Located
        # at upper right corner.
        self._coords_legend = CoordsLegend()
        self._coords_legend.setParentItem(self.plotItem.vb)
        self._coords_legend.anchor(itemPos=(1, 0),
                                   parentPos=(1, 0),
                                   offset=(-5, 5))

        # Picker legend for selected pixel coordinates.
        # Located at lower right corner.
        self._picker_legend = PickerLegend()
        self._picker_legend.setParentItem(self.plotItem.vb)
        self._picker_legend.anchor(itemPos=(1, 1),
                                   parentPos=(1, 1),
                                   offset=(-5, -5))

        # Connect image plot and item signals
        imageItem = plotItem.imageItem
        imageItem.sigImageChanged.connect(self.update)
        imageItem.clicked.connect(self.select_pixel)
        imageItem.hovered.connect(self.hover_pixel)
        plotItem.imageTransformed.connect(self._update_geometry)

    def activate(self, active):
        """Called upon clicking Picker toolbutton"""
        self._indicator_rect.setVisible(False)
        self._coords_legend.setVisible(False)

        # If there is previously selected pixel, show selection rect and
        # picker legend.
        if self._selected_pixel is not None:
            self._selection_rect.setVisible(active)
            self._picker_legend.setVisible(active)
            if active:
                self._update_geometry()
                self.update()

    @Slot(float, float)
    def select_pixel(self, x, y):
        if not self._selection_rect.isVisible():
            self._selection_rect.show()
            self._picker_legend.show()

        self._selection_rect.setPos(x, y)
        self.update()

    @Slot()
    def update(self):
        if not self._selection_rect.isVisible():
            return

        if not self._picker_legend.isVisible():
            self._picker_legend.show()

        x, y = np.floor(self._selection_rect.absolute_position).astype(int)
        self._selected_pixel = (x, y)

        value = self.plotItem.image[y, x]
        color = self.plotItem.imageItem.get_color(x, y)
        pos = self._selection_rect.pos()

        self._picker_legend.set_value(pos.x(), pos.y(), value, color)

    @Slot(object, object)
    def hover_pixel(self, x, y):
        if x is None and y is None:
            if self._indicator_rect.isVisible():
                self._indicator_rect.hide()
                self._coords_legend.hide()
        else:
            if not self._indicator_rect.isVisible():
                self._indicator_rect.show()
                self._coords_legend.show()
            self._indicator_rect.setPos(x, y)

        if x is not None:
            x = float_to_string(x)
        if y is not None:
            y = float_to_string(y)
        self._coords_legend.set_value(x, y)

    def _update_geometry(self):
        """Adjusts picker items geometry with the current image transform"""
        self._indicator_rect.update_geometry_from_transform(
            self.plotItem.axes_transform[TF_SCALING],
            self.plotItem.axes_transform[TF_TRANSLATION])
        self._selection_rect.update_geometry_from_transform(
            self.plotItem.axes_transform[TF_SCALING],
            self.plotItem.axes_transform[TF_TRANSLATION])

    def destroy(self):
        imageItem = self.plotItem.imageItem

        imageItem.sigImageChanged.disconnect(self.update)
        imageItem.clicked.disconnect(self.select_pixel)
        imageItem.hovered.disconnect(self.hover_pixel)
        self.plotItem.imageTransformed.disconnect(self._update_geometry)

        for item in [self._indicator_rect, self._selection_rect,
                     self._coords_legend, self._picker_legend]:
            self.plotItem.vb.removeItem(item)

            # Only instances of QObject have deleteLater(). It is said that
            # when an QGraphicsItem went out of scope, it is automatically
            # deleted. But we can use sip.delete() just to make sure? We can
            # also use multiple inheritance such as:
            # IndicatorRectangle(QGraphicsRectItem, QObject)
            if isinstance(item, QGraphicsObject):
                item.deleteLater()

        self.deleteLater()


class IndicatorRectangle(QGraphicsRectItem):
    """A graphics item that is used to indicate pixel position when the image
       is hovered or clicked."""

    def __init__(self):
        QGraphicsRectItem.__init__(self, 0, 0, 1, 1)

        pen = QPen(QColor(Qt.white))
        pen.setCosmetic(True)
        self.setPen(pen)
        self.hide()

        self._scaling = np.array([1, 1])
        self._translation = np.array([0, 0])

    def hoverEnterEvent(self, event):
        self.savedPen = self.pen()
        self.setPen(QPen(QColor(0, 0, 0)))
        super().hoverEnterEvent(event)

    def hoverLeaveEvent(self, event):
        self.setPen(self.savedPen)
        event.ignore()
        super().hoverLeaveEvent(event)

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            event.accept()
        else:
            event.ignore()
        super().mousePressEvent(event)

    def mouseMoveEvent(self, event):
        pass

    def update_geometry_from_transform(self, scaling, translation):
        x, y = (self.absolute_position * scaling) + translation
        w, h = self._absolute_size * scaling
        self._scaling = scaling
        self._translation = translation

        self.prepareGeometryChange()
        self.setRect(QRectF(0, 0, w, h))
        self.setPos(QPointF(x, y))

    @property
    def absolute_position(self):
        pos = self.pos()
        scaled = np.array([pos.x(), pos.y()]) - self._translation
        return scaled / self._scaling

    @property
    def _absolute_size(self):
        rect = self.rect()
        return np.array([rect.width(), rect.height()]) / self._scaling



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 30, 2011
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

from qtpy.QtCore import Slot
from qtpy.QtWidgets import QPushButton, QVBoxLayout, QWidget

from karabo.common.api import (
    KARABO_DAEMON_MANAGER, KARABO_SCHEMA_DEFAULT_SCENE)
from karabogui import messagebox
from karabogui.events import KaraboEvent, register_for_broadcasts
from karabogui.navigation.system_view import SystemTreeView
from karabogui.request import get_scene_from_server
from karabogui.singletons.api import get_config
from karabogui.topology.api import is_device_online

from .base import BasePanelWidget
from .tool_widget import SearchBar


class TopologyPanel(BasePanelWidget):
    def __init__(self):
        super().__init__("System Topology")
        event_map = {
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.AccessLevelChanged: self._event_access_level,
            KaraboEvent.ShowDevice: self._event_show_device
        }
        register_for_broadcasts(event_map)

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel."""
        widget = QWidget(self)
        main_layout = QVBoxLayout(widget)
        main_layout.setContentsMargins(2, 2, 2, 2)

        self.tree_view = SystemTreeView(widget)
        self.sbar = SearchBar(parent=widget)
        self.daemon_button = QPushButton("Service Manager", parent=widget)
        self.daemon_button.clicked.connect(self._retrieve_service_scene)
        self.daemon_button.setVisible(True)
        self.sbar.setView(self.tree_view)

        main_layout.addWidget(self.sbar)
        main_layout.addWidget(self.daemon_button)
        main_layout.addWidget(self.tree_view)

        return widget

    def __repr__(self):
        topic = get_config()["broker_topic"]
        return f"<TopologyPanel topic={topic}>"

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_access_level(self, data):
        self.sbar.reset(enable=True)

    def _event_network(self, data):
        status = data.get('status', False)
        self.sbar.reset(status)
        self.daemon_button.setVisible(status)
        if not status:
            self.close_popup_widget()

    def _event_show_device(self, data):
        if data.get('showTopology'):
            self.panel_container.switch_to_panel(self)

    # -----------------------------------------------------------------------

    def close_popup_widget(self):
        widget = self.tree_view.popupWidget
        if widget is not None:
            widget.close()

    @Slot()
    def _retrieve_service_scene(self):
        instance_id = KARABO_DAEMON_MANAGER
        if not is_device_online(instance_id):
            messagebox.show_error(
                "The `KaraboDaemonManager` device is not online.",
                parent=self)
            return
        get_scene_from_server(instance_id, KARABO_SCHEMA_DEFAULT_SCENE)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from collections import OrderedDict
from enum import Enum

from karabo.native import AccessLevel

ACCESS_LEVELS = OrderedDict()
ACCESS_LEVELS['Expert'] = AccessLevel.EXPERT
ACCESS_LEVELS['Operator'] = AccessLevel.OPERATOR
ACCESS_LEVELS['Observer'] = AccessLevel.OBSERVER

USERNAMES = [level.lower() for level in ACCESS_LEVELS.keys()]

# The globally defined access level variable which is verified and set by the
# GUI server eventually
GLOBAL_ACCESS_LEVEL = AccessLevel.OBSERVER
HIGHEST_ACCESS_LEVEL = AccessLevel.OBSERVER
ONE_TIME_TOKEN = None

AUTHENTICATION_SERVER = None

TEMPORARY_SESSION_USER = None
TEMPORARY_SESSION_WARNING = False
SESSION_END_NOTICE = False

SERVER_READ_ONLY = None

ACCESS_LEVEL_MAP = {
    "observer": 0,
    "operator": 1,
    "expert": 2,
    "god": 5}


class AccessRole(Enum):
    """The AccessRole Enum describes eventually validated action in the
    karabo GUI"""
    SCENE_EDIT = 'scene_edit'
    MACRO_EDIT = 'macro_edit'
    PROJECT_EDIT = 'project_edit'
    SERVICE_EDIT = 'service_edit'
    CONSOLE_EDIT = 'console_edit'
    INSTANCE_CONTROL = 'instance_control'
    APPLY_SCENE_EDIT = 'apply_edit'
    CONFIGURATION_DELETE = 'delete_configuration'


ACCESS_LEVEL_ROLES = {
    AccessRole.SCENE_EDIT: AccessLevel.EXPERT,
    AccessRole.MACRO_EDIT: AccessLevel.OPERATOR,
    AccessRole.PROJECT_EDIT: AccessLevel.OPERATOR,
    AccessRole.SERVICE_EDIT: AccessLevel.OPERATOR,
    AccessRole.CONSOLE_EDIT: AccessLevel.EXPERT,
    AccessRole.INSTANCE_CONTROL: AccessLevel.OPERATOR,
    AccessRole.CONFIGURATION_DELETE: AccessLevel.EXPERT,
    AccessRole.APPLY_SCENE_EDIT: AccessLevel.OPERATOR}


def access_role_allowed(role):
    """Return on runtime if action in the GUI are allowed"""
    return GLOBAL_ACCESS_LEVEL >= ACCESS_LEVEL_ROLES[role]


def is_authenticated():
    """one time token is set for only User-Authenticated login"""
    return ONE_TIME_TOKEN is not None


def is_end_of_session():
    return SESSION_END_NOTICE is True


def is_temporary_session() -> bool:
    """Check if session is temporary"""
    return TEMPORARY_SESSION_USER is not None


def reset_login():
    """Reset the global login information"""
    global ONE_TIME_TOKEN, TEMPORARY_SESSION_WARNING
    global TEMPORARY_SESSION_USER, SESSION_END_NOTICE, SERVER_READ_ONLY
    ONE_TIME_TOKEN = None
    TEMPORARY_SESSION_USER = None
    TEMPORARY_SESSION_WARNING = False
    SESSION_END_NOTICE = False
    SERVER_READ_ONLY = None


def get_access_level_for_role(role: AccessRole) -> str:
    """Get the access level needed for the AccessRole. """
    return ACCESS_LEVEL_ROLES[role].name


def set_read_only_mode(value: bool):
    global SERVER_READ_ONLY
    if SERVER_READ_ONLY is None:
        SERVER_READ_ONLY = value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QRect

from karabo.common.scenemodel import api as models
from karabogui.controllers.api import get_model_controller

from .const import QT_BOX_LAYOUT_DIRECTION
from .layout.api import BoxLayout, GridLayout, GroupLayout
from .shapes import ArrowShape, LineShape, PathShape, RectangleShape
from .widget.api import (
    ControllerContainer, DeviceSceneLinkWidget, ImageRendererWidget,
    InstanceStatusWidget, LabelWidget, PopupButtonWidget, SceneLinkWidget,
    StickerWidget, UnknownSvgWidget, UnknownWidget, WebLinkWidget)

_LAYOUT_CLASSES = (BoxLayout, GridLayout, GroupLayout)
_SHAPE_CLASSES = (ArrowShape, LineShape, PathShape, RectangleShape)
_WIDGET_CLASSES = (ControllerContainer, DeviceSceneLinkWidget, LabelWidget,
                   SceneLinkWidget, StickerWidget, PopupButtonWidget,
                   WebLinkWidget, UnknownSvgWidget, UnknownWidget,
                   ImageRendererWidget, InstanceStatusWidget)

_SCENE_OBJ_FACTORIES = {
    models.BoxLayoutModel: lambda m, p: BoxLayout(
        m, QT_BOX_LAYOUT_DIRECTION[m.direction]),
    models.FixedLayoutModel: lambda m, p: GroupLayout(m),
    models.GridLayoutModel: lambda m, p: GridLayout(m),
    models.LabelModel: LabelWidget,
    models.ArrowPolygonModel: lambda m, p: ArrowShape(model=m),
    models.LineModel: lambda m, p: LineShape(model=m),
    models.PathModel: lambda m, p: PathShape(model=m),
    models.PopupButtonModel: PopupButtonWidget,
    models.RectangleModel: lambda m, p: RectangleShape(model=m),
    models.SceneLinkModel: SceneLinkWidget,
    models.StickerModel: StickerWidget,
    models.WebLinkModel: WebLinkWidget,
    models.UnknownWidgetDataModel: UnknownWidget,
    models.DeviceSceneLinkModel: DeviceSceneLinkWidget,
    models.InstanceStatusModel: InstanceStatusWidget,
    models.ImageRendererModel: ImageRendererWidget,
    models.UnknownXMLDataModel: lambda m, p: UnknownSvgWidget.create(
        m, parent=p),
}


def find_top_level_model(scene_model, model):
    """Recursively find the model in the ``parent_model`` tree which matches
    the given ``model`` and return its parent model which is the top level
    model in the end.
    """

    def recurse(parent_model, seek_model):
        if isinstance(parent_model, models.BaseLayoutModel):
            for child in parent_model.children:
                result = recurse(child, seek_model)
                if result is not None:
                    return parent_model

        return parent_model if parent_model is seek_model else None

    for child in scene_model.children:
        top_level_model = recurse(child, model)
        if top_level_model is not None:
            return top_level_model


def replace_model_in_top_level_model(layout_model, parent_model, old_model,
                                     new_model):
    """Recursively find the given ``old_model`` in the model tree and replace
    it with the given ``new_model``.

    This method returns ``True`` if the model was replaced.
    """
    if isinstance(parent_model, models.BaseLayoutModel):
        for child in parent_model.children:
            result = replace_model_in_top_level_model(parent_model, child,
                                                      old_model, new_model)
            if result:
                return True
    elif parent_model is old_model:
        # Replace old model with new model
        layout_children = layout_model.children
        index = layout_children.index(parent_model)
        layout_children.remove(parent_model)
        layout_children.insert(index, new_model)
        # Enforce recalculation of geometry
        layout_model.width = 0
        layout_model.height = 0
        return True
    return False


def iter_widgets_and_models(obj, model, object_dict):
    """Given a scene object hierarchy rooted in `obj`, return all of the
    widgets within, and their models.
    """
    if is_widget(obj):
        yield obj, model
    elif is_layout(obj):
        for child_model in model.children:
            child_obj = object_dict.get(child_model)
            yield from iter_widgets_and_models(child_obj, child_model,
                                               object_dict)


def add_object_to_layout(obj, layout):
    """Add a SceneView object to a layout"""
    if is_shape(obj):
        layout._add_shape(obj)
    elif is_widget(obj):
        layout._add_widget(obj)
    elif is_layout(obj):
        layout._add_layout(obj)
    # Make sure the object is showing
    obj.show()


def remove_object_from_layout(obj, layout, object_dict):
    """Remove a SceneView object from a layout"""
    if is_shape(obj):
        layout._remove_shape(obj)
    elif is_widget(obj):
        layout._remove_widget(obj)
    elif is_layout(obj):
        layout._remove_layout(obj)
        # Recursively remove all children
        for model in obj.model.children:
            child_obj = object_dict.get(model)
            if child_obj:
                remove_object_from_layout(child_obj, obj, object_dict)

    # Hide object from scene until reparenting is done
    obj.hide()


def create_object_from_model(layout, model, parent_widget, object_dict,
                             scene_visible):
    """Create a SceneView object to mirror a data model object"""
    obj = object_dict.get(model)
    if obj is None:
        factory = _SCENE_OBJ_FACTORIES.get(model.__class__)
        if factory:
            obj = factory(model, parent_widget)
        else:
            klass = get_model_controller(model)
            if klass is None:
                # This widget model is either deprecated/removed or new.
                obj = UnknownWidget(model, parent_widget)
            else:
                obj = ControllerContainer(klass, model, parent_widget)

    # Add the new scene object to the layout
    if obj is not None:
        if model not in object_dict:
            object_dict[model] = obj
        add_object_to_layout(obj, layout)
        if is_layout(obj):
            # recurse
            fill_root_layout(obj, model, parent_widget, object_dict,
                             scene_visible)
            model_rect = QRect(model.x, model.y, model.width, model.height)
            if model_rect.isEmpty():
                # Ask the layout to calculate a suitable size
                obj.invalidate()
                rect = obj.geometry()
                if rect.isEmpty():
                    model_rect.setSize(obj.sizeHint())
                else:
                    model_rect = rect
            obj.set_geometry(model_rect)
        elif is_widget(obj):
            model_rect = QRect(model.x, model.y, model.width, model.height)
            widget_rect = obj.geometry()
            if model_rect.isEmpty():
                model_rect.setSize(obj.sizeHint())
                obj.set_geometry(model_rect)
            elif not widget_rect.isEmpty() and widget_rect != model_rect:
                # Handle the case when widgets unparented from a layout have
                # geometry which disagrees with their model!
                obj.set_geometry(widget_rect)
            # Don't just call set_visible(scene_visible) here!
            # The visibility count will be wrong if scene_visible is False!
            if scene_visible:
                obj.set_visible(True)


def fill_root_layout(layout, parent_model, parent_widget, object_dict,
                     scene_visible):
    """Recursively build scene GUI objects for a given parent model object.
    Whenever a layout is encountered, its children are then added recursively.

    `object_dict` is a cache of already created GUI objects.
    """
    for child_model in parent_model.children:
        create_object_from_model(layout, child_model, parent_widget,
                                 object_dict, scene_visible)


def bring_object_to_front(obj):
    if is_widget(obj):
        obj.raise_()
    elif is_layout(obj):
        for child in obj:
            layout = child.layout()
            if layout is not None:
                bring_object_to_front(layout)
                continue
            widget = child.widget()
            if widget is not None:
                bring_object_to_front(widget)


def send_object_to_back(obj):
    if is_widget(obj):
        obj.lower()
    elif is_layout(obj):
        for child in obj:
            layout = child.layout()
            if layout is not None:
                send_object_to_back(layout)
                continue
            widget = child.widget()
            if widget is not None:
                send_object_to_back(widget)


def is_layout(scene_obj):
    """Returns True if `scene_obj` is a layout"""
    return isinstance(scene_obj, _LAYOUT_CLASSES)


def is_shape(scene_obj):
    """Returns True if `scene_obj` is a shape"""
    return isinstance(scene_obj, _SHAPE_CLASSES)


def is_rect(scene_obj):
    return isinstance(scene_obj, RectangleShape)


def is_widget(scene_obj):
    """Returns True if `scene_obj` is a widget"""
    return isinstance(scene_obj, _WIDGET_CLASSES)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabogui.graph.common.enums import MouseTool, ROITool
from karabogui.testing import GuiTestCase

from ..controller import ToolbarController
from ..toolsets import MouseModeToolset


class TestToolbarController(GuiTestCase):

    def setUp(self):
        super().setUp()
        self.toolbar = ToolbarController()

    def test_basics(self):
        # Check if toolset contains only MouseMode
        assert len(self.toolbar.toolsets) == 1
        assert MouseTool in self.toolbar.toolsets

        # Check if default mouse mode toolset contains three buttons
        controller = self.toolbar.toolsets[MouseTool]
        assert isinstance(controller, MouseModeToolset)
        buttons = controller.buttons
        assert len(buttons) == 3
        assert set(buttons.keys()) == {MouseTool.Pointer, MouseTool.Zoom,
                                       MouseTool.Move}

        # Check if toolbar contains the buttons
        actions = self.toolbar.widget.actions()
        toolbar_buttons = [action._toolbar_button for action in actions]
        assert len(toolbar_buttons) == 3
        assert set(toolbar_buttons) == set(controller.buttons.values())

        # Check mouse mode buttons default checked state
        assert buttons[MouseTool.Pointer].isChecked()
        assert not buttons[MouseTool.Zoom].isChecked()
        assert not buttons[MouseTool.Move].isChecked()

    def test_add_toolset(self):
        widget = self.toolbar.widget

        # Add existing toolset
        self.toolbar.add_toolset(MouseTool)
        assert len(self.toolbar.toolsets) == 1
        # Count the number of buttons in the toolbar
        mouse_mode = self.toolbar.toolsets[MouseTool]
        assert len(widget.actions()) == len(mouse_mode.buttons)
        assert len(self.toolbar.buttons) == 3
        assert "Pointer" in self.toolbar.buttons
        assert "Zoom" in self.toolbar.buttons
        assert "Move" in self.toolbar.buttons

        # Add new toolset
        self.toolbar.add_toolset(ROITool)
        assert len(self.toolbar.toolsets) == 2
        roi = self.toolbar.toolsets[ROITool]
        # Count the number of buttons in the toolbar, including one separator
        num_actions = len(mouse_mode.buttons) + len(roi.buttons) + 1
        assert len(widget.actions()) == num_actions
        assert len(self.toolbar.buttons) == 5
        assert "Crosshair" in self.toolbar.buttons
        assert "Rect" in self.toolbar.buttons

        # Add invalid toolset
        self.toolbar.add_toolset("foo")
        # Check if there is no changes.
        assert len(self.toolbar.toolsets) == 2
        assert len(widget.actions()) == num_actions
        assert len(self.toolbar.buttons) == 5

    def test_add_optional_tool(self):
        # Add optional tool to toolbar with MouseMode toolset
        self.toolbar.add_tool(MouseTool.Picker)
        assert len(self.toolbar.toolsets) == 1
        # Count the number of buttons in the toolbar
        controller = self.toolbar.toolsets[MouseTool]
        assert isinstance(controller, MouseModeToolset)
        buttons = controller.buttons
        assert len(buttons) == 4
        assert set(buttons.keys()) == {MouseTool.Pointer, MouseTool.Zoom,
                                       MouseTool.Move, MouseTool.Picker}
        assert "Picker" in self.toolbar.buttons
        assert "Pointer" in self.toolbar.buttons
        assert "Zoom" in self.toolbar.buttons
        assert "Move" in self.toolbar.buttons



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest import main, mock

from qtpy.QtCore import QItemSelectionModel, Qt

from karabo.native import Hash, Timestamp
from karabogui.dialogs.api import LogDialog
from karabogui.events import KaraboEvent
from karabogui.singletons.mediator import Mediator
from karabogui.testing import GuiTestCase, singletons

data = [
    Hash("timestamp", Timestamp().toLocal(),
         "category", "XHQ_EG_DG/MOTOR/MOTOR1",
         "type", "INFO",
         "traceback", "",
         "message", "Schema updated"),

    Hash("timestamp", Timestamp().toLocal(),
         "category", "XHQ_EG_DG/MOTOR/MOTOR1",
         "type", "ERROR",
         "traceback", "",
         "message", "Error limit switch")
]


class TestLoginDialog(GuiTestCase):

    def test_basic_dialog(self):
        """Test the basic of the log dialog"""
        mediator = Mediator()
        with singletons(mediator=mediator):
            server_id = "swerver"
            dialog = LogDialog(server_id, parent=None)
            assert dialog.server_id == server_id
            assert dialog.ui_server_id.text() == server_id

            logs = Hash("content", data)
            dialog.request_handler(True, logs)
            table_view = dialog.log_widget.table
            table_model = table_view.model()
            assert table_model.rowCount() == 2

            # Get instanceId
            index = table_model.index(0, 2)
            actual_instanceId = "XHQ_EG_DG/MOTOR/MOTOR1"
            instanceId = table_model.data(index, role=Qt.DisplayRole)
            assert instanceId == actual_instanceId

            # Select that index
            table_view.selectionModel().setCurrentIndex(
                index, QItemSelectionModel.ClearAndSelect)
            path = "karabogui.widgets.log.broadcast_event"
            with mock.patch(path) as broad:
                dialog.log_widget.onItemDoubleClicked(index)
                broad.assert_called_with(KaraboEvent.ShowDevice,
                                         {"deviceId": actual_instanceId})

            path = "karabogui.dialogs.log_dialog.call_device_slot"
            with mock.patch(path) as call:
                self.click(dialog.ui_request)
                call.assert_called_once()


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import argparse
import sys

from karabo.common.scenemodel.api import set_scene_reader
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.programs.base import create_gui_app, init_gui
from karabogui.singletons.api import get_config


def run_gui(ns):
    app = create_gui_app([])

    set_scene_reader(lazy=True)
    # Set the development mode
    get_config()["development"] = ns.dev

    # some final initialization
    init_gui(app, use_splash=True)

    # Make the main window
    broadcast_event(KaraboEvent.CreateMainWindow, {})

    # then start the event loop
    app.exec()
    app.deleteLater()
    sys.exit()


def main():
    ap = argparse.ArgumentParser(description='Karabo GUI')
    ap.add_argument('-dev', '--dev', action='store_true')
    run_gui(ap.parse_args())


if __name__ == '__main__':
    main()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################

import numpy as np
from qtpy.QtWidgets import QAction
from traits.api import Instance, WeakRef

from karabo.common.scenemodel.api import (
    VectorHistGraphModel, build_model_config)
from karabogui import icons
from karabogui.binding.api import NDArrayBinding, VectorNumberBinding
from karabogui.controllers.api import (
    BaseBindingController, get_array_data, register_binding_controller)
from karabogui.graph.plots.api import HistogramDialog, KaraboPlotView


@register_binding_controller(
    ui_name='Vector HistoGram Graph',
    klassname='VectorHistGraph',
    binding_type=(VectorNumberBinding, NDArrayBinding),
    priority=-10,
    can_show_nothing=False)
class VectorHistogramGraph(BaseBindingController):
    """The controller for display of data in a histogram format"""
    model = Instance(VectorHistGraphModel, args=())

    _plot = WeakRef(object)

    def create_widget(self, parent):
        widget = KaraboPlotView(parent=parent)
        widget.add_cross_target()
        widget.add_toolbar()
        widget.enable_export()
        widget.stateChanged.connect(self._change_model)
        self._plot = widget.add_curve_item()

        widget.restore(build_model_config(self.model))

        hist_action = QAction("Histogram settings ...", widget)
        hist_action.triggered.connect(self.configure_histogram)
        hist_action.setIcon(icons.histogram)
        widget.addAction(hist_action)

        return widget

    # ----------------------------------------------------------------
    # Qt Slots

    def _change_model(self, content):
        self.model.trait_set(**content)

    def configure_histogram(self):
        content, ok = HistogramDialog.get(build_model_config(self.model),
                                          parent=self.widget)
        if ok:
            self.model.trait_set(**content)
            if len(self.proxy.value):
                self.value_update(self.proxy)

    def _plot_histogram(self, value):
        """The plot function for the histogram called by `value_update`"""
        # XXX: Do we still need this?
        finite = np.isfinite(value)
        if not np.all(finite):
            value = value.copy()
            value[~finite] = 0

        if self.model.auto:
            start, stop = value.min(), value.max()
        else:
            start = min([self.model.start, self.model.stop])
            stop = max([self.model.start, self.model.stop])
        bins = np.linspace(start=start, stop=stop, num=self.model.bins + 1,
                           endpoint=True)
        hist, edges = np.histogram(value, bins=bins)
        if len(edges) > 1:
            bin_w = (stop - start) / (self.model.bins)
            self._plot.setData(edges - bin_w / 2, hist, fillLevel=0,
                               stepMode="center")

    def value_update(self, proxy):
        value, _ = get_array_data(proxy, default=[])
        if not len(value):
            self._plot.setData([], stepMode=None)
            return

        # NOTE: We cast boolean as int!
        if value.dtype == bool:
            value = value.astype(int)

        self._plot_histogram(value)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt

from karabo.common.api import State
from karabo.common.scenemodel.api import LineEditModel
from karabo.native import Char, Configurable, String
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..textedit import EditableLineEdit


class Object(Configurable):
    state = String(defaultValue=State.ON)
    prop = String(allowedStates={State.ON})


class ObjectChar(Configurable):
    prop = Char()


def test_editable_line_edit(gui_app):
    # set up
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    controller = EditableLineEdit(proxy=proxy, model=LineEditModel())
    controller.create(None)

    # test focus policy
    assert controller.widget.focusPolicy() == Qt.StrongFocus

    # test set value
    controller._last_cursor_pos = 0
    set_proxy_value(proxy, "prop", "Blah")
    assert controller.widget.text() == "Blah"
    assert controller.widget.cursorPosition() == 0

    # test edit value
    controller.widget.textChanged.emit("Wha??")
    assert proxy.edit_value == "Wha??"

    # access level
    assert controller.widget.isEnabled()
    set_proxy_value(proxy, "state", "ERROR")
    assert not controller.widget.isEnabled()

    # enabled
    controller.setEnabled(True)
    assert controller.widget.isEnabled()
    controller.setEnabled(False)
    assert not controller.widget.isEnabled()

    # teardown
    controller.destroy()
    assert controller.widget is None


def test_char_editable_line_edit(gui_app):
    # setup
    proxy = get_class_property_proxy(ObjectChar.getClassSchema(), "prop")
    controller = EditableLineEdit(proxy=proxy, model=LineEditModel())
    controller.create(None)

    # test set value
    controller._last_cursor_pos = 0
    set_proxy_value(proxy, "prop", "D")
    assert controller.widget.text() == "D"
    assert controller.widget.cursorPosition() == 0

    # test edit value
    set_proxy_value(proxy, "prop", "$")
    assert controller.widget.text() == "$"
    controller.widget.setText("12345")
    # stop at the first char
    assert proxy.edit_value == "1"

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from unittest.mock import Mock

from traits.api import Undefined

from karabo.native import Hash
from karabogui.binding.api import (
    DeviceClassProxy, DeviceProxy, ImageBinding, PropertyProxy, ProxyStatus,
    apply_default_configuration, build_binding, extract_sparse_configurations,
    get_editor_value)
from karabogui.testing import (
    assert_trait_change, get_all_props_schema, get_class_property_proxy,
    get_pipeline_schema, get_pipeline_vector_schema, get_simple_schema,
    get_slotted_schema, singletons)


def test_device_proxy_classes():
    schema = get_simple_schema()
    binding = build_binding(schema)

    for factory in (DeviceProxy, DeviceClassProxy):
        proxy = factory(binding=binding)

        prop = proxy.get_property_binding('foo')
        assert prop is binding.value.foo

        prop = proxy['foo']
        assert prop is binding.value.foo

        not_there = proxy.get_property_binding('nope.blah')
        assert not_there is None

        assert proxy['nope.blah'] is None


def test_device_class_proxy_status():
    topology = Mock()
    topology.get_attributes.return_value = None
    with singletons(topology=topology):
        schema = get_simple_schema()
        binding = build_binding(schema)
        proxy = DeviceClassProxy(server_id='swerver', binding=binding)
        assert repr(proxy) == ("<DeviceClassProxy classId=Simple "
                               "serverId=swerver>")

        assert proxy.status == ProxyStatus.NOSERVER
        topology.get_attributes.assert_called_with('server.swerver')

        topology.get_attributes.return_value = {}
        proxy = DeviceClassProxy(binding=binding)
        assert proxy.status == ProxyStatus.NOPLUGIN

        topology.get_attributes.return_value = {'deviceClasses': ['Simple']}
        proxy = DeviceClassProxy(binding=binding)
        assert proxy.status == ProxyStatus.OFFLINE

        proxy.status = ProxyStatus.REQUESTED
        proxy.schema_update = True
        assert proxy.status == ProxyStatus.OFFLINE
        assert not proxy.online


def test_device_class_proxy_refresh():
    network = Mock()
    with singletons(network=network):
        schema = get_simple_schema()
        binding = build_binding(schema)
        proxy = DeviceClassProxy(server_id='swerver', binding=binding)
        proxy.refresh_schema()

        assert proxy.status == ProxyStatus.REQUESTED
        network.onGetClassSchema.assert_called_with('swerver', 'Simple')


def test_device_proxy_slots():
    network = Mock()
    with singletons(network=network):
        schema = get_slotted_schema()
        binding = build_binding(schema)
        apply_default_configuration(binding)

        proxy = DeviceProxy(device_id='dev', server_id='swerver',
                            binding=binding)
        slot_proxy = PropertyProxy(path='callme', root_proxy=proxy)
        slot_proxy.execute()
        network.onExecute.assert_called_with('dev', 'callme', False)


def test_device_proxy_status():
    network = Mock()
    with singletons(network=network):
        schema = get_simple_schema()
        binding = build_binding(schema)
        proxy = DeviceProxy(device_id='dev', server_id='swerver',
                            binding=binding)

        assert repr(proxy) == "<DeviceProxy deviceId=dev>"
        assert proxy.status == ProxyStatus.OFFLINE
        proxy.add_monitor()
        proxy.status = ProxyStatus.ONLINE
        network.onGetDeviceSchema.assert_called_with('dev')

        proxy.schema_update = True
        network.onStartMonitoringDevice.assert_called_with('dev')
        assert proxy.status == ProxyStatus.SCHEMA

        proxy.config_update = True
        assert proxy.status == ProxyStatus.MONITORING

        proxy.remove_monitor()
        network.onStopMonitoringDevice.assert_called_with('dev')
        assert proxy.status == ProxyStatus.ONLINE

        # Full cycle, reset mock
        network.reset_mock()

        proxy.add_monitor()
        assert proxy.status == ProxyStatus.ONLINEREQUESTED
        network.onGetDeviceSchema.assert_called_with('dev')
        proxy.schema_update = True
        assert proxy.status == ProxyStatus.SCHEMA
        proxy.config_update = True
        assert proxy.status == ProxyStatus.MONITORING

        proxy.remove_monitor()
        proxy.status = ProxyStatus.ONLINE
        proxy.add_monitor()
        assert proxy.status == ProxyStatus.ONLINEREQUESTED


def test_device_proxy_data_feeds():
    network = Mock()
    with singletons(network=network):
        schema = get_simple_schema()
        binding = build_binding(schema)
        proxy = DeviceProxy(device_id='dev', server_id='swerver',
                            binding=binding)

        proxy.connect_pipeline('pipe')
        network.onSubscribeToOutput.assert_called_with('dev', 'pipe', True)

        proxy.disconnect_pipeline('pipe')
        network.onSubscribeToOutput.assert_called_with('dev', 'pipe', False)

        historic_data = None

        def _event_fired(value):
            nonlocal historic_data
            historic_data = value

        apply_default_configuration(proxy.binding)
        binding = proxy.get_property_binding('foo')
        assert binding.value  # defaults to True

        binding.on_trait_change(_event_fired, 'historic_data')
        proxy.publish_historic_data('foo', False)
        assert historic_data is False


def test_property_proxy():
    schema = get_simple_schema()
    binding = build_binding(schema)
    root_proxy = DeviceProxy(device_id='dev', binding=binding)
    proxy = PropertyProxy(root_proxy=root_proxy, path='bar')
    other = PropertyProxy(root_proxy=root_proxy, path='foo')

    assert proxy.binding is binding.value.bar
    assert other.binding is binding.value.foo

    assert proxy.key == 'dev.bar'
    assert other.key == 'dev.foo'

    other.path = 'nope'
    assert other.binding is None

    assert other != proxy
    other.path = 'bar'
    assert other == proxy


def test_property_proxy_device_value():
    schema = get_simple_schema()
    binding = build_binding(schema)

    # First a device proxy
    root_proxy = DeviceProxy(device_id='dev', binding=binding)
    proxy = PropertyProxy(root_proxy=root_proxy, path='bar')
    proxy.value = 'Remote'
    assert repr(proxy) == "<PropertyProxy key=dev.bar>"

    assert proxy.get_device_value() == 'Remote'

    # Then a class proxy
    root_proxy = DeviceClassProxy(binding=binding)
    proxy = PropertyProxy(root_proxy=root_proxy, path='foo')
    assert proxy.get_device_value()  # defaultValue of 'foo' is True


def test_property_proxy_existing():
    """Test the existing property of a property proxy"""
    network = Mock()
    with singletons(network=network):
        schema = get_simple_schema()
        binding = build_binding(schema)
        root_proxy = DeviceProxy(device_id='dev', server_id='swerver',
                                 binding=binding)
        proxy = PropertyProxy(root_proxy=root_proxy, path='bar')
        other = PropertyProxy(root_proxy=root_proxy, path='fooo')

        assert root_proxy.status == ProxyStatus.OFFLINE
        assert proxy.existing
        assert other.existing
        root_proxy.add_monitor()
        root_proxy.status = ProxyStatus.ONLINE
        assert proxy.existing
        assert other.existing
        network.onGetDeviceSchema.assert_called_with('dev')

        root_proxy.schema_update = True
        network.onStartMonitoringDevice.assert_called_with('dev')
        assert root_proxy.status == ProxyStatus.SCHEMA
        assert proxy.existing
        assert proxy.binding is not None
        assert not other.existing
        assert other.binding is None


def test_property_proxy_history():
    network = Mock()
    with singletons(network=network):
        schema = get_simple_schema()
        binding = build_binding(schema)
        root_proxy = DeviceProxy(device_id='dev', binding=binding)
        proxy = PropertyProxy(root_proxy=root_proxy, path='foo')

        proxy.get_history('start_time', 'end_time', max_value_count=1)
        network.onGetPropertyHistory.assert_called_with(
            'dev', 'foo', 'start_time', 'end_time', 1)


def test_property_proxy_edit_values():
    schema = get_all_props_schema()
    proxy = get_class_property_proxy(schema, 'n')

    proxy.value = 5
    assert get_editor_value(proxy) == 5

    proxy.edit_value = 42
    assert get_editor_value(proxy) == 42

    proxy.revert_edit()
    assert get_editor_value(proxy) == 5
    assert proxy.edit_value is None


def test_property_proxy_pipeline():
    schema = get_pipeline_schema()
    binding = build_binding(schema)
    root_proxy = DeviceProxy(binding=binding)
    proxy = PropertyProxy(root_proxy=root_proxy, path='output.data.image')

    assert isinstance(proxy.binding, ImageBinding)
    assert proxy.pipeline_parent_path == 'output'

    proxy = PropertyProxy(root_proxy=root_proxy,
                          path='output.data.vector')
    assert proxy.binding is None
    # No pipeline established
    assert proxy.pipeline_parent_path == ''


def test_property_proxy_pipeline_monitoring():
    network = Mock()
    with singletons(network=network):
        schema = get_pipeline_schema()
        binding = build_binding(schema)
        root_proxy = DeviceProxy(device_id='dev', binding=binding)
        root_proxy.status = ProxyStatus.ONLINE
        proxy = PropertyProxy(root_proxy=root_proxy, path='output.data.image')

        assert not proxy.visible

        proxy.start_monitoring()
        assert proxy.visible
        assert proxy.root_proxy._pipeline_subscriptions['output'] == 1
        network.onSubscribeToOutput.assert_called_with('dev', 'output', True)

        proxy.stop_monitoring()
        assert not proxy.visible
        network.onSubscribeToOutput.assert_called_with('dev', 'output', False)
        assert proxy.root_proxy._pipeline_subscriptions['output'] == 0

        # Test that we don't subscribe to a property that is not available
        network.reset_mock()
        proxy = PropertyProxy(root_proxy=root_proxy,
                              path='output.data.vector')
        assert proxy.binding is None
        assert not proxy.visible
        proxy.start_monitoring()
        assert proxy.root_proxy.status is ProxyStatus.ONLINEREQUESTED
        assert proxy.visible
        assert proxy.root_proxy._monitor_count > 0
        # not called
        network.onSubscribeToOutput.assert_not_called()

        # Got the schema now
        schema = get_pipeline_vector_schema()
        with assert_trait_change(proxy, 'pipeline_parent_path'):
            build_binding(schema, existing=binding)
            assert proxy.binding is not None

        assert proxy.pipeline_parent_path == 'output'
        assert proxy.root_proxy._pipeline_subscriptions['output'] == 1
        assert proxy.root_proxy.status is ProxyStatus.SCHEMA
        network.onSubscribeToOutput.assert_called_with('dev', 'output', True)


def test_schema_updates():
    schema = get_simple_schema()
    binding = build_binding(schema)
    proxy = DeviceProxy(binding=binding)

    with assert_trait_change(proxy, 'schema_update'):
        build_binding(schema, existing=proxy.binding)


def test_multi_device_config_extraction():
    schema = get_all_props_schema()
    dev_one = DeviceProxy(device_id='one', binding=build_binding(schema))
    dev_two = DeviceProxy(device_id='two', binding=build_binding(schema))
    prop_bool = PropertyProxy(root_proxy=dev_one, path='a')
    prop_string = PropertyProxy(root_proxy=dev_two, path='m')

    # faulty list of nodes binding
    prop_lon = PropertyProxy(root_proxy=dev_two, path='j1')
    assert prop_lon.binding is None

    prop_bool.edit_value = True
    prop_string.edit_value = 'yo'
    node_val = Hash('_NodeOne', Hash('zero', 'hello'))
    prop_lon.edit_value = [node_val]
    configs = extract_sparse_configurations([prop_bool, prop_string, prop_lon])

    assert 'one' in configs and 'two' in configs
    assert configs['one']['a'] is True
    assert configs['two']['m'] == 'yo'


def test_delegation_with_schema_update():
    schema = get_simple_schema()
    binding = build_binding(schema)
    device = DeviceProxy(binding=binding)
    proxy = PropertyProxy(root_proxy=device, path='bar')

    with assert_trait_change(proxy, 'binding:value'):
        proxy.value = 'Hello'

    with assert_trait_change(proxy, 'binding'):
        build_binding(schema, existing=device.binding)

    # The value on the proxy has been reset to the default for strings
    assert proxy.value is Undefined



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import importlib
import importlib.metadata
import os

from traits.api import Undefined

import karabogui.access as krb_access
from karabo.common.api import KARABO_SCHEMA_REGEX


def axis_label(proxy):
    """Return the axis label for a PropertyProxy instance
    """
    binding = proxy.binding
    if binding is None:
        return ''

    unit = binding.unit_label
    name = binding.displayedName
    return f"{name} [{unit}]" if unit else name


def get_class_const_trait(klass, name):
    """Return the value of a `Constant` trait which has been added to a class
    object by `register_binding_controller`.
    """
    trait = klass.class_traits()[name]
    return trait.default


def has_options(binding):
    """Returns True if a binding has any `options` defined."""
    return len(binding.options) > 0


def has_regex(binding):
    """Returns True if the binding has any `regex` defined."""
    return binding.attributes.get(KARABO_SCHEMA_REGEX) is not None


def get_regex(binding, default=None):
    """Returns the `regex` of a binding, if not present takes the `default`"""
    return binding.attributes.get(KARABO_SCHEMA_REGEX, default)


def populate_controller_registry():
    """Make sure all the `BaseBindingController` subclasses are registered.

    Yup. gui_registry_loader lives on! Hopefully a little less ugly in this
    form?

    NOTE: This will not work if the GUI is ever packaged as compiled modules
    (ie: if the .py files are not present).
    """
    ROOT_PACKAGE = 'karabogui.controllers'
    SUBPACKAGES = ('display', 'edit')
    for pkg in SUBPACKAGES:
        pkg = ROOT_PACKAGE + '.' + pkg
        module = importlib.import_module(pkg)
        # Only the Python files
        mod_files = [fn for fn in os.listdir(os.path.dirname(module.__file__))
                     if fn.endswith('.py')]
        for mod_fname in mod_files:
            submodule = os.path.splitext(mod_fname)[0]
            importlib.import_module(f'{pkg}.{submodule}')

    load_extensions()


def load_extensions():
    """Load the entrypoints for our gui extensions"""
    EXT_PACKAGE = 'karabogui.gui_extensions'

    all_entry_points = importlib.metadata.entry_points()
    entry_points = all_entry_points.select(group=EXT_PACKAGE)
    for entry in entry_points:
        try:
            entry.load()
        except Exception as e:
            print(f'Cannot load plugin {entry.name}: {e}')


def with_display_type(displayType):
    """Create a checker function for the `is_compatible` argument of
    `register_binding_controller` which looks for a specific display type
    of property.
    """

    def is_compatible(binding):
        dt = binding.displayType
        return dt == displayType

    return is_compatible


def is_proxy_allowed(proxy):
    """Retrieve if the reconfiguration of the ``PropertyProxy`` is allowed
    """
    root_proxy = proxy.root_proxy
    value = root_proxy.state_binding.value
    if value is Undefined or not value:
        return False

    binding = proxy.binding
    is_allowed = binding.is_allowed(value)
    is_accessible = (krb_access.GLOBAL_ACCESS_LEVEL >=
                     binding.requiredAccessLevel)
    return is_accessible and is_allowed



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on October 27, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op
import re
from functools import partial
from io import StringIO

from qtpy.QtWidgets import QAction, QDialog, QMenu, QMessageBox
from traits.api import Instance

from karabo.common.api import BaseSavableModel, walk_traits_object
from karabo.common.project.api import get_user_cache
from karabo.common.scenemodel.api import SceneModel, read_scene, write_scene
from karabo.native import read_project_model
from karabogui import icons, messagebox
from karabogui.access import AccessRole, access_role_allowed
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.itemtypes import ProjectItemTypes
from karabogui.project.dialog.object_handle import (
    ObjectDuplicateDialog, ObjectEditDialog)
from karabogui.singletons.api import (
    get_config, get_db_conn, get_panel_wrangler)
from karabogui.util import getOpenFileName, getSaveFileName, move_to_cursor
from karabogui.wizards.api import CinemaWizardController

from .bases import BaseProjectController, ProjectControllerUiData


class SceneController(BaseProjectController):
    """ A controller for SceneModel objects
    """
    # Redefine model with the correct type
    model = Instance(SceneModel)

    def context_menu(self, project_controller, parent=None):
        scene_allowed = access_role_allowed(AccessRole.SCENE_EDIT)

        menu = QMenu(parent)
        edit_action = QAction(icons.edit, 'Edit', menu)
        edit_action.triggered.connect(partial(self._edit_scene, parent=parent))
        edit_action.setEnabled(scene_allowed)

        dupe_action = QAction(icons.editCopy, 'Duplicate', menu)
        dupe_action.triggered.connect(partial(self._duplicate_scene,
                                              project_controller,
                                              parent=parent))
        dupe_action.setEnabled(scene_allowed)

        delete_action = QAction(icons.delete, 'Delete', menu)
        delete_action.triggered.connect(partial(self._delete_scene,
                                                project_controller,
                                                parent=parent))
        delete_action.setEnabled(scene_allowed)

        save_as_action = QAction(icons.saveAs, 'Save to file', menu)
        save_as_action.triggered.connect(partial(self._save_scene_to_file,
                                                 parent=parent))
        replace_action = QAction(icons.editPasteReplace,
                                 'Replace from file', menu)
        replace_action.triggered.connect(partial(self._replace_scene,
                                                 project_controller,
                                                 parent=parent))
        replace_action.setEnabled(scene_allowed)

        revert_action = QAction(icons.revert, 'Revert changes', menu)
        revert_action.triggered.connect(partial(self._revert_changes,
                                                parent=parent))
        can_revert = not self._is_showing() and self.model.modified
        revert_action.setEnabled(can_revert and scene_allowed)

        project_model = project_controller.model
        cinema_action = QAction(icons.run, 'Create cinema link', menu)
        cinema_action.triggered.connect(partial(self._create_cinema_link,
                                                project_model=project_model,
                                                parent=parent))

        menu.addAction(edit_action)
        menu.addAction(dupe_action)
        menu.addAction(delete_action)
        menu.addSeparator()
        menu.addAction(save_as_action)
        menu.addAction(replace_action)
        menu.addAction(revert_action)
        menu.addSeparator()
        menu.addAction(cinema_action)

        return menu

    def info(self):
        return {'type': ProjectItemTypes.SCENE,
                'uuid': self.model.uuid,
                'simple_name': self.model.simple_name}

    def create_ui_data(self):
        return ProjectControllerUiData(icon=icons.image)

    def double_click(self, project_controller, parent=None):
        broadcast_event(KaraboEvent.ShowSceneView, {'model': self.model})

    def delete_press(self, project_controller, parent=None):
        """Reimplemented function on `BaseProjectController`"""
        self._delete_scene(project_controller, parent)

    # ----------------------------------------------------------------------
    # action handlers

    def _replace_scene(self, project_controller, parent=None):
        """ Replace a scene from local disk
        """
        # Make sure scene is closed before!
        scene = self.model
        path = get_config()['data_dir']
        directory = path if path and op.isdir(path) else ""

        fn = getOpenFileName(caption='Replace scene',
                             filter='SVG Files (*.svg)',
                             directory=directory,
                             parent=parent)
        if not fn:
            return

        # Store scene dir path
        get_config()['data_dir'] = op.dirname(fn)

        project = project_controller.model
        # NOTE: Close the old one first!
        broadcast_event(KaraboEvent.RemoveProjectModelViews,
                        {'models': [scene]})

        if scene in project.scenes:
            # Read SceneModel for information to replace with! We catch all
            # errors because we can have wrong user input.
            scene.assure_svg_data()

            try:
                new_scene = read_scene(fn)
                new_scene.assure_svg_data()
            except Exception:
                messagebox.show_error("Scene file could not be read!",
                                      parent=parent)
                return

            # Only use copyable traits
            scene.copy_traits(new_scene, traits=['width', 'height',
                                                 'children'])

            def visitor(model):
                if isinstance(model, BaseSavableModel):
                    model.initialized = True

            # Mark everything as initialized, otherwise `modified` won't work.
            walk_traits_object(scene, visitor)

            scene.modified = True

    def _delete_scene(self, project_controller, parent=None):
        """ Remove the scene associated with this item from its project
        """
        scene = self.model
        ask = ('Are you sure you want to delete \"<b>{}</b>\".<br /> '
               'Continue action?'.format(scene.simple_name))
        msg_box = QMessageBox(QMessageBox.Question, 'Delete scene',
                              ask, QMessageBox.Yes | QMessageBox.No,
                              parent=parent)
        msg_box.setModal(False)
        msg_box.setDefaultButton(QMessageBox.No)
        move_to_cursor(msg_box)
        if msg_box.exec() == QMessageBox.Yes:
            project = project_controller.model
            if scene in project.scenes:
                project.scenes.remove(scene)

            broadcast_event(KaraboEvent.RemoveProjectModelViews,
                            {'models': [scene]})

    def _edit_scene(self, parent=None):
        dialog = ObjectEditDialog(object_type='scene', model=self.model,
                                  parent=parent)
        move_to_cursor(dialog)
        result = dialog.exec()
        if result == QDialog.Accepted:
            self.model.assure_svg_data()
            self.model.simple_name = dialog.simple_name

    def _duplicate_scene(self, project_controller, parent=None):
        scene = self.model
        project = project_controller.model
        dialog = ObjectDuplicateDialog(scene.simple_name, parent=parent)
        move_to_cursor(dialog)
        if dialog.exec() == QDialog.Accepted:
            scene.assure_svg_data()
            xml = write_scene(scene)
            for simple_name in dialog.duplicate_names:
                dupe_scene = read_scene(StringIO(xml))
                dupe_scene.assure_svg_data()
                dupe_scene.simple_name = simple_name
                dupe_scene.reset_uuid()
                project.scenes.append(dupe_scene)

    def _is_showing(self):
        scene = self.model
        wrangler = get_panel_wrangler()
        return wrangler.is_showing_project_item(scene)

    def _revert_changes(self, parent=None):
        scene = self.model
        # NOTE: The domain only changes when a project is saved or loaded
        domain = get_db_conn().default_domain
        data = get_user_cache().retrieve(domain, scene.uuid)
        if data is None:
            msg = 'That scene has been removed from the local cache!'
            messagebox.show_error(msg, title='Revert failed', parent=parent)
            return

        read_project_model(StringIO(data), existing=scene)
        scene.modified = False

    def _save_scene_to_file(self, parent=None):
        config = get_config()
        path = config['data_dir']
        directory = path if path and op.isdir(path) else ""
        scene = self.model
        filename = scene.simple_name
        filename = re.sub(r'[\W]', '-', filename)
        fn = getSaveFileName(caption='Save scene to file',
                             filter='SVG (*.svg)',
                             suffix='svg',
                             selectFile=filename,
                             directory=directory,
                             parent=parent)
        if not fn:
            return

        # Store old scene dialog path
        config['data_dir'] = op.dirname(fn)

        if not fn.endswith('.svg'):
            fn = f'{fn}.svg'

        scene.assure_svg_data()
        with open(fn, 'w') as fout:
            fout.write(write_scene(scene))

    def _create_cinema_link(self, project_model=None, parent=None):
        wizard = CinemaWizardController(selected_scenes=[self.model],
                                        project_model=project_model,
                                        parent=parent)
        wizard.run()



#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on May 31, 2013
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from functools import partial

from qtpy.QtCore import QPoint, Qt, Slot
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import (
    QAbstractItemView, QAction, QDialog, QHeaderView, QInputDialog, QMenu,
    QTreeView)

from karabo.common.api import KARABO_DAEMON_MANAGER, Capabilities
from karabo.native import Timestamp
from karabogui import icons, messagebox
from karabogui.access import (
    AccessRole, access_role_allowed, get_access_level_for_role)
from karabogui.const import TOOLTIP_INIT_CONFIG, TOOLTIP_RUNTIME_CONFIG
from karabogui.dialogs.api import (
    ConfigurationFromPastDialog, DeviceCapabilityDialog,
    InitConfigurationDialog)
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.itemtypes import NavigationItemTypes
from karabogui.navigation.system_filter_model import TopologyFilterModel
from karabogui.request import call_device_slot, get_scene_from_server
from karabogui.singletons.api import (
    get_manager, get_network, get_selection_tracker)
from karabogui.topology.api import is_device_online
from karabogui.util import (
    get_reason_parts, load_configuration_from_file, move_to_cursor,
    save_configuration_to_file)
from karabogui.widgets.popup import PopupWidget

from .system_model import SystemTreeModel
from .tools import DeviceSceneHandler, ServerLogHandler

DISABLED_TOOLTIP = (
    f"Requires minimum '"
    f"{get_access_level_for_role(AccessRole.SERVICE_EDIT)}' access level.")


class SystemTreeView(QTreeView):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setObjectName("systemTreeView")
        self._selected_proxy = None  # A BaseDeviceProxy

        model = SystemTreeModel(parent=self)
        proxy_model = TopologyFilterModel(parent=self,
                                          source_model=model)
        self.setModel(proxy_model)
        self.setSelectionModel(proxy_model.selectionModel)

        header = self.header()
        header.moveSection(1, 0)
        header.setSectionResizeMode(0, QHeaderView.Stretch)
        header.setSectionResizeMode(1, QHeaderView.Fixed)
        header.setStretchLastSection(False)
        icon_size = 26
        header.setMaximumSectionSize(icon_size)
        header.resizeSection(1, icon_size)

        proxy_model.setFilterKeyColumn(0)
        proxy_model.signalItemChanged.connect(self.onSelectionChanged)
        proxy_model.modelReset.connect(self.resetExpand)

        self.setSelectionMode(QAbstractItemView.SingleSelection)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)

        self._setupContextMenu()
        self.customContextMenuRequested.connect(
            self.onCustomContextMenuRequested)
        self.setDragEnabled(True)

        self.handler_list = [DeviceSceneHandler(), ServerLogHandler()]

        self.expanded = False
        self.popupWidget = None
        self.header().sectionDoubleClicked.connect(self.onDoubleClickHeader)
        self.setUniformRowHeights(True)

    def _setupContextMenu(self):
        self.setContextMenuPolicy(Qt.CustomContextMenu)

        # Device server instance menu
        self.mHostItem = QMenu(self)
        text = "Network information"
        self.acNetworkInfo = QAction(icons.about, text, self)
        self.acNetworkInfo.setStatusTip(text)
        self.acNetworkInfo.setToolTip(text)
        self.acNetworkInfo.triggered.connect(self.onNetworkInfo)
        self.mHostItem.addAction(self.acNetworkInfo)

        text = "About"
        self.acAbout = QAction(icons.about, "About", self)
        self.acAbout.setStatusTip(text)
        self.acAbout.setToolTip(text)
        self.acAbout.triggered.connect(self.onAbout)

        text = "Time information"
        self.acTimeInformation = QAction(icons.clock, text, self)
        self.acTimeInformation.triggered.connect(self.onTimeInformation)
        self.acTimeInformation.setVisible(False)  # Classes don't have time

        # Device server instance menu
        self.mServerItem = QMenu(self)
        self.mServerItem.setToolTipsVisible(True)

        text = "Set logger level"
        self.acLoggerLevel = QAction(icons.edit, text, self)
        self.acLoggerLevel.setStatusTip(text)
        self.acLoggerLevel.setToolTip(text)
        self.acLoggerLevel.triggered.connect(self.onLoggerLevel)
        self.mServerItem.addAction(self.acLoggerLevel)

        text = "Shutdown server"
        self.acKillServer = QAction(icons.delete, text, self)
        self.acKillServer.setStatusTip(text)
        self.acKillServer.setToolTip(text)
        self.acKillServer.triggered.connect(self.onKillInstance)
        self.mServerItem.addAction(self.acKillServer)

        self.mServerItem.addSeparator()
        self.mServerItem.addAction(self.acAbout)

        self.mServerItem.addAction(self.acTimeInformation)

        # Device class/instance menu
        self.mDeviceItem = QMenu(self)
        self.mDeviceItem.setToolTipsVisible(True)

        text = "Open configuration (*.xml)"
        self.acOpenFromFile = QAction(icons.load, text, self)
        self.acOpenFromFile.setStatusTip(text)
        self.acOpenFromFile.setToolTip(text)
        self.acOpenFromFile.triggered.connect(self.onOpenFromFile)
        self.mDeviceItem.addAction(self.acOpenFromFile)

        text = "Save configuration as (*.xml)"
        self.acSaveToFile = QAction(icons.saveAs, text, self)
        self.acSaveToFile.setStatusTip(text)
        self.acSaveToFile.setToolTip(text)
        self.acSaveToFile.triggered.connect(self.onSaveToFile)
        self.mDeviceItem.addAction(self.acSaveToFile)

        self.mDeviceItem.addSeparator()

        text = "Runtime configuration"
        self.ac_config_past = QAction(icons.clock, text, self)
        self.ac_config_past.setStatusTip(text)
        self.ac_config_past.setToolTip(TOOLTIP_RUNTIME_CONFIG)
        self.ac_config_past.triggered.connect(self.onGetConfigurationFromPast)
        self.mDeviceItem.addAction(self.ac_config_past)

        text = "Init configuration"
        self.ac_config_name = QAction(icons.initconfig, text, self)
        self.ac_config_name.setStatusTip(text)
        self.ac_config_name.setToolTip(TOOLTIP_INIT_CONFIG)
        self.ac_config_name.triggered.connect(self.onGetInitConfiguration)
        self.mDeviceItem.addAction(self.ac_config_name)

        self.mDeviceItem.addSeparator()

        text = "Shutdown device"
        self.acKillDevice = QAction(icons.delete, text, self)
        self.acKillDevice.setStatusTip(text)
        self.acKillDevice.setToolTip(text)
        self.acKillDevice.triggered.connect(self.onKillInstance)
        self.mDeviceItem.addAction(self.acKillDevice)

        text = "Open device scene..."
        self.acOpenScene = QAction(text, self)
        self.acOpenScene.setStatusTip(text)
        self.acOpenScene.setToolTip(text)
        self.acOpenScene.setVisible(False)  # Most items don't have scenes
        self.acOpenScene.triggered.connect(self.onOpenDeviceScene)
        self.mDeviceItem.addAction(self.acOpenScene)

        self.mDeviceItem.addSeparator()

        self.mDeviceItem.addAction(self.acTimeInformation)
        self.mDeviceItem.addAction(self.acAbout)

    def currentIndex(self):
        return self.model().currentIndex()

    def indexInfo(self, index=None):
        """Return the info about the index.

        Defaults to the current index if index is None.
        """
        if index is None:
            index = self.currentIndex()
        return self.model().indexInfo(index)

    def clear(self):
        self._selected_proxy = None
        self.clearSelection()
        self.model().clear()

    def scrollTo(self, index, hint=QAbstractItemView.EnsureVisible):
        """Reimplementation of the Qt function
        """
        # XXX: We used to expand the index here!
        super().scrollTo(index, hint)

    # ----------------------------
    # Events

    def mouseDoubleClickEvent(self, event):
        index = self.currentIndex()
        node = self.model().index_ref(index)
        # QModelIndexes tend to outlive the node objects which they reference!
        if node is None:
            return
        info = node.info()
        for handler in self.handler_list:
            if handler.can_handle(info):
                handler.handle(info, parent=self)
                event.accept()
                return

        super().mouseDoubleClickEvent(event)

    # ----------------------------
    # Slots

    @Slot()
    def onNetworkInfo(self):
        if not is_device_online(KARABO_DAEMON_MANAGER):
            messagebox.show_error(
                "The KaraboDaemonManager is not online ...", parent=self)
            return

        index = self.currentIndex()
        node = self.model().index_ref(index)
        if node is None:
            return

        def show_network_info(success, reply):
            if not success:
                reason, details = get_reason_parts(reply)
                messagebox.show_error(reason, details=details, parent=self)
                return

            if self.popupWidget is None:
                self.popupWidget = PopupWidget(parent=self)

            payload = reply["payload"]
            network = payload["network"]
            self.popupWidget.setInfo(network)
            self.popupWidget.adjustSize()
            move_to_cursor(self.popupWidget)
            self.popupWidget.show()

        host = node.info()["hostId"]
        call_device_slot(
            show_network_info, KARABO_DAEMON_MANAGER, "requestNetwork",
            host=host)

    @Slot()
    def onAbout(self):
        index = self.currentIndex()
        node = self.model().index_ref(index)
        if node is None:
            return
        if self.popupWidget is None:
            self.popupWidget = PopupWidget(parent=self)
        self.popupWidget.setInfo(node.attributes)

        pos = QCursor.pos()
        pos.setX(pos.x() + 10)
        pos.setY(pos.y() + 10)
        self.popupWidget.move(pos)
        self.popupWidget.show()

    @Slot()
    def onGetConfigurationFromPast(self):
        info = self.indexInfo()
        device_id = info.get('deviceId')
        dialog = ConfigurationFromPastDialog(instance_id=device_id,
                                             parent=self)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    @Slot()
    def onGetInitConfiguration(self):
        info = self.indexInfo()
        device_id = info.get('deviceId')
        dialog = InitConfigurationDialog(instance_id=device_id,
                                         parent=self)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    @Slot()
    def onTimeInformation(self):
        """Get the time information from a `server` or `device`"""
        info = self.indexInfo()
        if info is None:
            return

        def show_server_information(instanceId, success, reply):
            if not success:
                return

            if self.popupWidget is None:
                self.popupWidget = PopupWidget(parent=self)

            actual = Timestamp.fromHashAttributes(reply["time", ...])
            attrs = {"instanceId": instanceId,
                     "Actual Timestamp": actual,
                     "Actual Timing Id": actual.tid}

            # Note: New protocol additions since 2.11
            if "timeServerId" in reply:
                attrs.update({"timeServerId": reply["timeServerId"]})
            if "reference" in reply:
                reference = Timestamp.fromHashAttributes(
                    reply["reference", ...])
                attrs.update({"Last Tick Timestamp": reference,
                              "Last Tick Timing Id": reference.tid})

            self.popupWidget.setInfo(attrs)
            self.popupWidget.adjustSize()
            move_to_cursor(self.popupWidget)
            self.popupWidget.show()

        node_type = info.get('type')
        if node_type is NavigationItemTypes.DEVICE:
            instanceId = info.get('deviceId')
        elif node_type is NavigationItemTypes.SERVER:
            instanceId = info.get('serverId')

        handler = partial(show_server_information, instanceId)
        call_device_slot(handler, instanceId, 'slotGetTime')

    @Slot()
    def onLoggerLevel(self):
        info = self.indexInfo()
        levels = ["DEBUG", "INFO", "WARN", "ERROR"]
        index = levels.index(info.get("log"))
        level, ok = QInputDialog.getItem(self, "Set the logger level", "",
                                         levels, index, False)
        if ok:
            serverId = info.get('serverId')
            network = get_network()
            network.onSetLogLevel(serverId, level)

    @Slot()
    def onKillInstance(self):
        info = self.indexInfo()
        node_type = info.get('type')
        manager = get_manager()

        if node_type is NavigationItemTypes.DEVICE:
            deviceId = info.get('deviceId')
            manager.shutdownDevice(deviceId, parent=self)
        elif node_type is NavigationItemTypes.SERVER:
            serverId = info.get('serverId')
            manager.shutdownServer(serverId, parent=self)

    @Slot()
    def onOpenDeviceScene(self):
        info = self.indexInfo()
        dialog = DeviceCapabilityDialog(
            device_id=info.get('deviceId', ''),
            capability=Capabilities.PROVIDES_SCENES,
            parent=self)
        if dialog.exec() == QDialog.Accepted:
            device_id = dialog.device_id
            scene_name = dialog.capa_name
            get_scene_from_server(device_id, scene_name=scene_name)

    @Slot(QPoint)
    def onCustomContextMenuRequested(self, pos):
        def _test_mask(mask, bit):
            return (mask & bit) == bit

        info = self.indexInfo()
        node_type = info.get('type', NavigationItemTypes.UNDEFINED)
        # Killing services is access level dependent!
        allow_service_edit = access_role_allowed(AccessRole.SERVICE_EDIT)
        if node_type is NavigationItemTypes.HOST:
            instance_control = access_role_allowed(AccessRole.INSTANCE_CONTROL)
            self.acNetworkInfo.setEnabled(instance_control)
            self.mHostItem.exec(QCursor.pos())
        elif node_type is NavigationItemTypes.SERVER:
            self.acKillServer.setEnabled(allow_service_edit)
            self.acLoggerLevel.setEnabled(allow_service_edit)
            _set_tooltip(self.acKillServer, allow_service_edit)
            _set_tooltip(self.acLoggerLevel, allow_service_edit)
            self.acAbout.setVisible(True)
            self.acTimeInformation.setVisible(False)
            self.mServerItem.exec(QCursor.pos())
        elif node_type is NavigationItemTypes.DEVICE:
            self.acKillDevice.setVisible(True)
            self.acKillDevice.setEnabled(allow_service_edit)
            _set_tooltip(self.acKillDevice, allow_service_edit)
            self.acAbout.setVisible(True)
            self.acTimeInformation.setVisible(True)
            has_scenes = _test_mask(info.get('capabilities', 0),
                                    Capabilities.PROVIDES_SCENES)
            self.acOpenScene.setVisible(has_scenes)
            self.mDeviceItem.exec(QCursor.pos())

    @Slot(str, object)
    def onSelectionChanged(self, item_type, proxy):
        """Called by the data model when an item is selected
        """
        self._selected_proxy = proxy

        # Grab control of the global selection
        get_selection_tracker().grab_selection(self.model().selectionModel)

        # Tell the configurator
        if item_type not in ('class', 'device'):
            # servers and hosts clear the configurator
            proxy = None
        broadcast_event(KaraboEvent.ShowConfiguration, {'proxy': proxy})

    @Slot()
    def onOpenFromFile(self):
        if self._selected_proxy is not None:
            load_configuration_from_file(self._selected_proxy, parent=self)

    @Slot()
    def onSaveToFile(self):
        if self._selected_proxy is not None:
            save_configuration_to_file(self._selected_proxy, parent=self)

    @Slot()
    def onDoubleClickHeader(self):
        if self.expanded:
            self.collapseAll()
        else:
            self.expandAll()

        index = self.model().currentIndex()
        if index.isValid():
            self.scrollTo(index)

    @Slot()
    def resetExpand(self):
        self.expandAll()

    def collapseAll(self):
        self.expanded = False
        super().collapseAll()

    def expandAll(self):
        self.expanded = True
        super().expandAll()


def _set_tooltip(action: QAction, enabled: bool) -> None:
    """Add tooltip about lacking acces level when menu-item is disabled."""
    text = action.text() if enabled else DISABLED_TOOLTIP
    action.setToolTip(text)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import re
from math import sqrt

from qtpy.QtCore import QPointF
from qtpy.QtGui import QPainterPath


class Parser:
    re = re.compile("([MmZzLlHhVvCcSsQqTtAa])|"
                    "(-?(([0-9]*[.][0-9]*)|[0-9]+)([eE][+-]?[0-9]*)?)")

    def next(self):
        if self.iter is None:
            self.token = ''
            raise StopIteration
        ret = float(self.number) if self.number is not None else self.token
        try:
            self.token, self.number, _, _, _ = next(self.iter).groups()
        except StopIteration:
            self.iter = None
        return ret

    def __init__(self, s):
        self.iter = self.re.finditer(s)

    def __iter__(self):
        while self.token is None:
            yield self.next()

    def parse(self):
        self.path = QPainterPath()
        self.pos = QPointF(0, 0)
        try:
            self.number = self.token = None
            self.next()
            while True:
                token = self.next()
                getattr(self, token.lower())(token.islower())
                self.lasttoken = token
        except (RuntimeError, StopIteration):
            # RuntimeError in generators since Python 3.8
            return self.path

    def points(self, relative):
        for x, y in zip(self, self):
            if relative:
                yield self.path.currentPosition() + QPointF(x, y)
            else:
                yield QPointF(x, y)

    def m(self, relative):
        p = next(self.points(relative))
        self.path.moveTo(p)
        self.l(relative)

    def z(self, relative):
        self.path.closeSubpath()

    def l(self, relative):  # noqa
        for p in self.points(relative):
            self.path.lineTo(p)

    def h(self, relative):
        for x in self:
            pos = self.path.currentPosition()
            if relative:
                pos += QPointF(x, 0)
            else:
                pos.setX(x)
            self.path.lineTo(pos)

    def v(self, relative):
        for y in self:
            pos = self.path.currentPosition()
            if relative:
                pos += QPointF(0, y)
            else:
                pos.setY(y)
            self.path.lineTo(pos)

    def c(self, relative):
        p = self.points(relative)
        for a, b, c in zip(p, p, p):
            self.path.cubicTo(a, b, c)
            self.lastcontrol = b

    def s(self, relative):
        p = self.points(relative)
        if self.lasttoken not in 'CcSs':
            self.lastcontrol = self.pos
        a = 2 * self.pos - self.lastcontrol
        for b, c in zip(p, p):
            self.path.cubicTo(a, b, c)
            a = 2 * c - b
            self.lastcontrol = b

    def q(self, relative):
        p = self.points(relative)
        for a, b in zip(p, p):
            self.path.quadTo(a, b)
            self.lastcontrol = a

    def t(self, relative):
        if self.lasttoken not in 'QqTt':
            self.lastcontrol = self.pos
        a = 2 * self.pos - self.lastcontrol
        for b in self.points(relative):
            self.path.quadTo(a, b)
            a = 2 * b - a
            self.lastcontrol = b

    def a(self, relative):
        c = None
        for rx, ry, phi, fa, fs, x, y in zip(*(self,) * 7):
            pp = (self.path.currentPosition() - QPointF(x, y)) / 2
            cp = (-1) ** (fa == fs) * sqrt(
                ((rx * ry) ** 2 - (rx * pp.y()) ** 2 - (ry * pp.x()) ** 2) /
                ((rx * pp.y()) ** 2 + (ry * pp.x()) ** 2)) * (
                     QPointF(rx * pp.y() / ry, -ry * pp.x() / rx))
            c = cp + (self.path.currentPosition() + QPointF(x, y)) / 2
        if c:
            self.lastcontrol = c



import argparse
import sys

import yaml
from qtpy.QtCore import Slot

import karabogui.const as global_constants
from karabo.common.scenemodel.api import SceneTargetWindow
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.messagebox import show_error
from karabogui.programs.base import create_gui_app, init_gui
from karabogui.programs.utils import close_app
from karabogui.singletons.api import get_db_conn, get_network, get_topology


def create_concert(data: dict):
    app = create_gui_app(sys.argv)
    init_gui(app, use_splash=True)

    def trigger_scenes():
        topology.system_tree.on_trait_change(
            trigger_scenes, "initialized", remove=True)
        domain = data["domain"]
        get_db_conn().default_domain = domain

        for scene in data["scenes"]:
            uuid = scene["uuid"]
            position = [int(scene["x"]), int(scene["y"])]
            db_scene = {"name": f"{domain}-{uuid}",
                        "target": uuid,
                        "position": position,
                        "target_window": SceneTargetWindow.Dialog
                        }
            broadcast_event(KaraboEvent.OpenSceneLink, db_scene)

    global_constants.APPLICATION_MODE = True
    topology = get_topology()
    # Attach to the topology
    topology.system_tree.on_trait_change(
        trigger_scenes, "initialized")
    network = get_network()

    @Slot()
    def _connect_handler(status):
        """Connect handler for direct connect attempts without dialog"""
        network.signalServerConnectionChanged.disconnect(_connect_handler)
        if not status:
            show_error("Error, could not connect to gui server "
                       f"<b>{data['host']}:{data['port']}</b>. Ending "
                       f"Concert.")
            topology.system_tree.on_trait_change(
                trigger_scenes, "initialized", remove=True)
            close_app()

    # We might want to connect directly to the gui server if all are provided
    if "host" in data and "port" in data:
        network.signalServerConnectionChanged.connect(_connect_handler)
        success = network.connectToServerDirectly(
            hostname=data["host"],
            port=data["port"])
    else:
        # Connect to the GUI Server via dialog"""
        success = network.connectToServer()
    return success, app


def run_concert(ns):
    """The concert is meant to load a yaml file for scene and network
    configuration

    Similar to the cinema, the scenes are spawned, but with positions.
    All scenes have the name ProjectDB|SceneName and are not editable!
    """
    with open(ns.yaml_file) as fh:
        data = yaml.safe_load(fh)
    success, app = create_concert(data)

    if success:
        app.exec()
        app.deleteLater()
        sys.exit()
    else:
        close_app()


help_message = """Path to yaml file. Example of the file content.

domain: CAS_INTERNAL
scenes:
  - uuid: '9d26d07b-d79e-4f71-99f1-469344da89ed'
    x: 1458
    y: 93
  - uuid: '9b73efec-4523-4a47-b88f-90edc2e9ed89'
    x: 2332
    y: 361
host: exflqr46957
port: 44444
"""


def main():
    arguments = argparse.ArgumentParser(
        description="Karabo Concert",
        formatter_class=argparse.RawTextHelpFormatter)
    arguments.add_argument("yaml_file", type=str, help=help_message)
    run_concert(arguments.parse_args())


if __name__ == "__main__":
    main()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QDialog, QInputDialog
from traits.api import Instance, String, on_trait_change

from karabo.common.scenemodel.api import FloatSpinBoxModel
from karabogui.binding.api import FloatBinding, get_editor_value, get_min_max
from karabogui.controllers.api import (
    BaseBindingController, is_proxy_allowed, register_binding_controller)
from karabogui.dialogs.api import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.util import (
    MouseWheelEventBlocker, SignalBlocker, generateObjectName)
from karabogui.widgets.api import DoubleSpinBox


@register_binding_controller(ui_name="Double SpinBox", can_edit=True,
                             klassname="FloatSpinBox",
                             binding_type=FloatBinding)
class FloatSpinBox(BaseBindingController):
    model = Instance(FloatSpinBoxModel, args=())
    _blocker = Instance(MouseWheelEventBlocker)
    _style_sheet = String

    def create_widget(self, parent):
        widget = DoubleSpinBox(parent)
        widget.setDecimals(self.model.decimals)
        widget.setSingleStep(self.model.step)
        widget.valueChanged[float].connect(self._on_user_edit)
        self._blocker = MouseWheelEventBlocker(widget)
        widget.installEventFilter(self._blocker)

        objectName = generateObjectName(self)
        widget.setObjectName(objectName)

        self._style_sheet = (f"QDoubleSpinBox#{objectName}" +
                             " {{ font: {}; font-size: {}pt; }}")
        # add actions
        step_action = QAction("Change Step...", widget)
        step_action.triggered.connect(self._change_step)
        widget.addAction(step_action)

        decimal_action = QAction("Change Decimals...", widget)
        decimal_action.triggered.connect(self._change_decimals)
        widget.addAction(decimal_action)

        format_action = QAction("Format field...", widget)
        format_action.triggered.connect(self._format_field)
        widget.addAction(format_action)

        # Apply initial formats
        self._apply_format(widget)

        return widget

    def binding_update(self, proxy):
        binding = proxy.binding
        label = None
        if binding is not None:
            label = f" {binding.unit_label}" if binding.unit_label else None
        low, high = get_min_max(binding)
        with SignalBlocker(self.widget):
            self.widget.setSuffix(label)
            self.widget.setRange(low, high)

    def value_update(self, proxy):
        value = get_editor_value(proxy)
        if value is not None:
            with SignalBlocker(self.widget):
                self.widget.setValue(value)

    def state_update(self, proxy):
        enable = is_proxy_allowed(proxy)
        self.widget.setEnabled(enable)

    @on_trait_change("model:decimals", post_init=True)
    def _set_decimals(self, value):
        if self.widget is not None:
            self.widget.setDecimals(value)

    @on_trait_change("model:step", post_init=True)
    def _set_step(self, value):
        if self.widget is not None:
            self.widget.setSingleStep(value)

    def _change_step(self):
        step = self.widget.singleStep()
        step, ok = QInputDialog.getDouble(
            self.widget, "Single Step", "Enter size of a single step", step)
        if ok:
            self.model.step = step

    def _change_decimals(self):
        # Override the starting value with something from the binding
        decimals = self.model.decimals
        decimals, ok = QInputDialog.getInt(
            self.widget, "Decimals", "Enter number of decimals",
            value=decimals, min=0, max=15)
        if ok:
            self.model.decimals = decimals

    def _on_user_edit(self, value):
        if self.proxy.binding is None:
            return
        self.proxy.edit_value = value

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        if widget is None:
            widget = self.widget

        # Apply font formatting
        sheet = self._style_sheet.format(
            self.model.font_weight,
            get_font_size_from_dpi(self.model.font_size))
        widget.setStyleSheet(sheet)
        widget.update()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog

from karabo.native import Configurable, VectorString
from karabogui.binding.api import apply_default_configuration
from karabogui.const import IS_MAC_SYSTEM
from karabogui.dialogs.listedit import ListEditDialog
from karabogui.testing import get_class_property_proxy

from ..strlist import EditableListElement


class Object(Configurable):
    prop = VectorString(defaultValue=["hi there"])


class ListEditMock(ListEditDialog):
    @property
    def values(self):
        return ["foo", "bar"]

    def exec(self):
        return QDialog.Accepted


def test_strlist_basics(gui_app, mocker):
    # setup
    proxy = get_class_property_proxy(Object.getClassSchema(), "prop")
    controller = EditableListElement(proxy=proxy)
    controller.create(None)
    apply_default_configuration(proxy.root_proxy.binding)\

    # focus policy
    if IS_MAC_SYSTEM:
        assert controller.widget.focusPolicy() == Qt.TabFocus
    else:
        assert controller.widget.focusPolicy() == Qt.StrongFocus

    # edit dialog
    target = "karabogui.controllers.edit.strlist.ListEditDialog"
    mocker.patch(target, new=ListEditMock)
    controller._on_edit_clicked()
    assert proxy.edit_value == ["foo", "bar"]

    # teardown
    controller.destroy()
    assert controller.widget is None



import pytest
from qtpy.QtCore import QEvent, QPoint, QSize, Qt
from qtpy.QtGui import QMouseEvent

from karabo.common.scenemodel.api import PopupButtonModel
from karabogui.sceneview.widget.api import PopupButtonWidget
from karabogui.widgets.api import TextPopupWidget

TEXT = """
Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod
tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At
vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd
gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet."""


@pytest.fixture()
def poup_button_widget(gui_app):
    model = PopupButtonModel(
        text=TEXT,  x=0, y=0, width=100, height=100)
    widget = PopupButtonWidget(model=model)
    return widget


def test_button_widget(poup_button_widget):
    """Test the popup sticker widget"""

    model = poup_button_widget.model
    widget = poup_button_widget

    size = QSize(model.width, model.height)
    # Check that the size is set correctly and is fixed.
    assert widget.size() == size


def test_popup_widget(poup_button_widget):
    """Verify that the popup widget appears with correc text on mouse click."""
    event = QMouseEvent(QEvent.MouseButtonPress, QPoint(0, 0), Qt.LeftButton,
                        Qt.LeftButton, Qt.NoModifier)
    poup_button_widget.mousePressEvent(event)

    poup_button_widget = poup_button_widget.children()[-1]
    assert poup_button_widget.isVisibleTo(poup_button_widget)
    assert isinstance(poup_button_widget, TextPopupWidget)
    assert poup_button_widget.text_edit.toPlainText() == TEXT
